import numpy as np

from pyHalo.PresetModels.wdm import WDM
from pyHalo.concentration_models import preset_concentration_models
from pyHalo.pyhalo import pyHalo
from pyHalo.realization_extensions import RealizationExtensions
from pyHalo.utilities import MinHaloMassULDM, de_broglie_wavelength


def ULDM(z_lens, z_source, log10_m_uldm, log10_fluc_amplitude=-0.8, fluctuation_size_scale=0.05,
          fluctuation_size_dispersion=0.2, n_fluc_scale=1.0, velocity_scale=200, sigma_sub=0.025,
         log10_sigma_sub=None, log_mlow=6., log_mhigh=10.,
        mass_function_model_subhalos='SHMF_SCHIVE2016', kwargs_mass_function_subhalos={},
        mass_function_model_fieldhalos='SCHIVE2016', kwargs_mass_function_fieldhalos={},
        concentration_model_subhalos='LAROCHE2022', kwargs_concentration_model_subhalos={},
        concentration_model_fieldhalos='LAROCHE2022', kwargs_concentration_model_fieldhalos={},
        truncation_model_subhalos='TRUNCATION_GALACTICUS', kwargs_truncation_model_subhalos={},
        truncation_model_fieldhalos='TRUNCATION_RN', kwargs_truncation_model_fieldhalos={},
        shmf_log_slope=-1.9, cone_opening_angle_arcsec=6., log_m_host=13.3, r_tidal=0.25,
        LOS_normalization=1.0, geometry_type='DOUBLE_CONE', kwargs_cosmo=None,
         uldm_plaw=1 / 3, flucs=True, flucs_shape='aperture', flucs_args={}, n_cut=50000, r_ein=1.0):

    """
    This generates realizations of ultra-light dark matter (ULDM), including the ULDM halo mass function and halo density profiles,
    as well as density fluctuations in the main deflector halo.

    Similarly to WDMGeneral, the functional form of the subhalo mass function is the same as the field halo mass function.
    However, this model differs from WDMGeneral by creating halos which are composite ULDM + NFW density profiles.
    The ULDM particle mass and core radius-halo mass power law exponent must now be specified. For details regarding ULDM halos,
    see Schive et al. 2014 (https://arxiv.org/pdf/1407.7762.pdf). Equations (3) and (7) give the soliton density profile
    and core radius, respectively.

    The differential halo mass function is described by three parameters, see Schive et al. 2016
    (https://arxiv.org/pdf/1508.04621.pdf):

    1) log_m0 - the log10 value of the characteristic mass, or the scale where the ULDM mass function begins
    to deviate from CDM:

        m0 = (1.6*10**10) * (m22)**(-4/3)   [solar masses],

    where

        m22 = m_uldm / 10**22 eV.

    2) b_uldm - modifies the log slope of the ULDM mass function, analogous to b_wdm (see WDMLovell2020)

    3) c_uldm - modifies the log slope of the ULDM mass function, analogous to c_wdm (see WDMLovell2020)

    The parametrization for the mass function is:

        n_uldm / n_cdm = (1 + (m / m_0)^b_uldm) ^ c_uldm,

    where Schive et al. 2016 determined that

        (m_0,    b_uldm,    c_uldm) = ( (1.6*10**10) * (m22)**(-4/3),    1.1,    2.2)

    As for the concentration relative to CDM, there hasa not been a detailed study to date. Hoever, since the formation
    history and collapse time determines concentration, we can reasonably use a WDM concentration-mass relation
    such as the Lovell 2020 formula,
    i.e. simply c_wdm = c_uldm, with (c_scale, c_power) = (15, -0.3).
    or the Bose et al. formula with (c_scale, c_power) = (60, -0.17) and a very negligible redshift dependence.

    The default model was computed using the formalism presented by Schneider et al. (2015) using a ULDM power spectrum,
    and results in a sharper cutoff with (c_scale, c_power, c_power_inner) = (3.348, -0.489, 1.5460)

    The form for the ULDM concentration-mass relation turnover is (1 + c_scale * (mhm/m)^c_power_inner)^c_power

    Furthermore, Schive et al. 2014 (https://arxiv.org/pdf/1407.7762.pdf) found a redshift-dependent minimum ULDM halo mass
    given by

        M_min(z) = a^(-3/4) * (Zeta(z)/Zeta(0))^(1/4) * M_min,0     [solar masses]

    where a is the cosmic scale factor,

        M_min,0 = 4.4*10^7 * m22^(-3/2)     [solar masses]

    and

        Zeta(z) = (18*pi^2 + 82(Om(z) - 1) - 39(Om(z) - 1)^2) / Om(z),

    where Om(z) is the matter density parameter.

    :param z_lens: main deflector redshift
    :param z_source: source redshift
    :param log10_m_uldm: log base 10 of the ULDM particle mass
    :param log10_fluc_amplitude: log base 10 of the fluctuation amplitude in the host halo density profile
    :param fluctuation_size_scale: sets the size of the fluctuations in host halo in units of de-Broglie wavelength
    :param fluctuation_size_dispersion: sets the variance of the size of the fluctuations
    :param n_fluc_scale: rescales the number of fluctuations
    :param velocity_scale: the velocity scale of the host halo used to convert particle mass to de-Broglie wavelength
    :param sigma_sub: amplitude of the subhalo mass function at 10^8 solar masses in units [# of halos / kpc^2]
    :param log10_sigma_sub: optional setting of sigma_sub in log10-scale (useful for log-uniform priors); if this is specified
    it overwrites the value of sigma_sub
    :param log_mlow: log base 10 of the minimum halo mass to render
    :param log_mhigh: log base 10 of the maximum halo mass to render
    :param mass_function_model_subhalos: mass function model for subhalos, see mass_function_models.py for a list
    :param kwargs_mass_function_subhalos: keyword arguments for the mass function model
    :param mass_function_model_fieldhalos: mass function model for field halos, see mass_function_models.py for a list
    :param kwargs_mass_function_fieldhalos: keyword arguments for the mass function model
    :param concentration_model_subhalos: the concentration-mass relation applied to subhalos,
    see concentration_models.py for a complete list of available models
    :param kwargs_concentration_model_subhalos: keyword arguments for the subhalo MC relation
    NOTE: keyword args returned by the load_concentration_model override these keywords with duplicate arguments
    :param concentration_model_subhalos: the concentration-mass relation applied to field halos,
    see concentration_models.py for a complete list of available models
    :param kwargs_concentration_model_fieldhalos: keyword arguments for the field halo MC relation
    NOTE: keyword args returned by the load_concentration_model override these keywords with duplicate arguments
    :param truncation_model_subhalos: the truncation model applied to subhalos, see truncation_models for a complete list
    :param kwargs_truncation_model_subhalos: keyword arguments for the truncation model applied to subhalos
    :param truncation_model_fieldhalos: the truncation model applied to field halos, see truncation_models for a
    complete list
    :param kwargs_truncation_model_fieldhalos: keyword arguments for the truncation model applied to field halos
    :param shmf_log_slope: the logarithmic slope of the subhalo mass function pivoting around 10^8 M_sun
    :param cone_opening_angle_arcsec: the opening angle of the rendering volume in arcsec
    :param log_m_host: log base 10 of the host halo mass [M_sun]
    :param r_tidal: the core radius of the host halo in units of the host halo scale radius. Subhalos are distributed
    in 3D with a cored NFW profile with this core radius
    :param geometry_type: string that specifies the geometry of the rendering volume; options include
    DOUBLE_CONE, CONE, CYLINDER
    :param kwargs_cosmo: keyword arguments that specify the cosmology, see Cosmology/cosmology.py
    :param uldm_plaw: sets the exponent for the core size/halo mass relation in ULDM
    :param flucs: bool; turns of/off fluctuations in the host halo density profile
    :param flucs_shape: the geometry in which to render fluctuations, options include 'aperture', 'ring', 'ellipse'
    :param flucs_args: keyword arguments for the specified geometry, see realization_extensions
    :param n_cut: maximum number of fluctuations to render; amplitudes of individual fluctuations get suppressed by
    sqrt(n_cut / n_total) if n_total > n_cut, where n_total is the number of expected fluctuations given the de-Broglie
    wavelength and the area in which the fluctuations area rendered
    :param r_ein: The characteristic angular used to convert fluctuations in convergence to fluctuations in mass
    :return: a realization of halos and fluctuations modeled as Gaussians in ULDM
    """

    # constants
    m22 = 10**(log10_m_uldm + 22)
    log_m0 = np.log10(1.6e10 * m22**(-4/3))

    # FIRST WE CREATE AN INSTANCE OF PYHALO, WHICH SETS THE COSMOLOGY
    pyhalo = pyHalo(z_lens, z_source, kwargs_cosmo)

    #compute M_min as described in documentation
    log_m_min = MinHaloMassULDM(log10_m_uldm, pyhalo.astropy_cosmo, log_mlow)
    kwargs_density_profile = {}
    kwargs_density_profile['log10_m_uldm'] = log10_m_uldm
    kwargs_density_profile['scale_nfw'] = False
    kwargs_density_profile['uldm_plaw'] = uldm_plaw
    kwargs_wdm = {'z_lens': z_lens, 'z_source': z_source, 'log_mc': log_m0, 'sigma_sub': sigma_sub,
                  'log10_sigma_sub': log10_sigma_sub,
                  'log_mlow': log_m_min, 'log_mhigh': log_mhigh,
                  'mass_function_model_subhalos': mass_function_model_subhalos,
                  'kwargs_mass_function_subhalos': kwargs_mass_function_subhalos,
                  'mass_function_model_fieldhalos': mass_function_model_fieldhalos,
                  'kwargs_mass_function_fieldhalos': kwargs_mass_function_fieldhalos,
                  'concentration_model_subhalos': concentration_model_subhalos,
                  'kwargs_concentration_model_subhalos': kwargs_concentration_model_subhalos,
                  'concentration_model_fieldhalos': concentration_model_fieldhalos,
                  'kwargs_concentration_model_fieldhalos': kwargs_concentration_model_fieldhalos,
                  'truncation_model_subhalos': truncation_model_subhalos,
                  'kwargs_truncation_model_subhalos': kwargs_truncation_model_subhalos,
                  'truncation_model_fieldhalos': truncation_model_fieldhalos,
                  'kwargs_truncation_model_fieldhalos': kwargs_truncation_model_fieldhalos,
                  'shmf_log_slope': shmf_log_slope, 'cone_opening_angle_arcsec': cone_opening_angle_arcsec,
                  'log_m_host': log_m_host, 'r_tidal': r_tidal, 'LOS_normalization': LOS_normalization,
                  'geometry_type': geometry_type, 'kwargs_cosmo': kwargs_cosmo,
                  'mdef_subhalos': 'ULDM', 'mdef_field_halos': 'ULDM',
                  'kwargs_density_profile': kwargs_density_profile
                  }

    uldm_no_fluctuations = WDM(**kwargs_wdm)
    if flucs: # add fluctuations to realization
        ext = RealizationExtensions(uldm_no_fluctuations)
        lambda_dB = de_broglie_wavelength(log10_m_uldm, velocity_scale) # de Broglie wavelength in kpc

        if flucs_args=={}:
            raise Exception('Must specify fluctuation arguments, see realization_extensions.add_ULDM_fluctuations')

        a_fluc = 10 ** log10_fluc_amplitude
        m_psi = 10 ** log10_m_uldm

        zlens_ref, zsource_ref = 0.5, 2.0
        mhost_ref = 10**13.3
        rein_ref = 1.0
        r_perp_ref = rein_ref * uldm_no_fluctuations.lens_cosmo.cosmo.kpc_proper_per_asec(zlens_ref)

        model, _ = preset_concentration_models('DIEMERJOYCE19')
        concentration_model_for_host = model(pyhalo.astropy_cosmo)
        sigma_crit_ref = uldm_no_fluctuations.lens_cosmo.get_sigma_crit_lensing(zlens_ref, zsource_ref)
        c_host_ref = concentration_model_for_host.nfw_concentration(mhost_ref, z_lens)
        rhos_ref, rs_ref, _ = uldm_no_fluctuations.lens_cosmo.NFW_params_physical(mhost_ref, c_host_ref, zlens_ref)
        xref = r_perp_ref/rs_ref
        if xref < 1:
            Fxref = np.arctanh(np.sqrt(1 - xref ** 2)) / np.sqrt(1 - xref ** 2)
        else:
            Fxref = np.arctan(np.sqrt(-1 + xref ** 2)) / np.sqrt(-1 + xref ** 2)
        sigma_host_ref = 2 * rhos_ref * rs_ref * (1-Fxref)/(xref**2 - 1)

        r_perp = r_ein * uldm_no_fluctuations.lens_cosmo.cosmo.kpc_proper_per_asec(z_lens)
        sigma_crit = uldm_no_fluctuations.lens_cosmo.get_sigma_crit_lensing(z_lens, z_source)
        c_host = concentration_model_for_host.nfw_concentration(10**log_m_host, z_lens)
        rhos, rs, _ = uldm_no_fluctuations.lens_cosmo.NFW_params_physical(10**log_m_host, c_host, z_lens)
        x = r_perp / rs
        if x < 1:
            Fx = np.arctanh(np.sqrt(1 - x ** 2)) / np.sqrt(1 - x ** 2)
        else:
            Fx = np.arctan(np.sqrt(-1 + x ** 2)) / np.sqrt(-1 + x ** 2)
        sigma_host = 2 * rhos * rs * (1 - Fx) / (x ** 2 - 1)

        fluctuation_amplitude = a_fluc * (m_psi / 1e-22) ** -0.5 * \
                                (sigma_crit_ref/sigma_crit) * (sigma_host/sigma_host_ref)

        uldm_realization = ext.add_ULDM_fluctuations(de_Broglie_wavelength=lambda_dB,
                                fluctuation_amplitude=fluctuation_amplitude,
                                fluctuation_size=lambda_dB * fluctuation_size_scale,
                                fluctuation_size_variance=lambda_dB * fluctuation_size_scale * fluctuation_size_dispersion,
                                n_fluc_scale=n_fluc_scale,
                                shape=flucs_shape,
                                args=flucs_args,
                                n_cut=n_cut)
        return uldm_realization
    else:
        return uldm_no_fluctuations



from pyHalo.Halos.tidal_truncation import TruncationRN, TruncationRoche, \
    AdiabaticTidesTruncation, TruncationSplashBack, TruncateMeanDensity, \
    TruncationGalacticus, ConstantTruncationArcsec, TruncationGalacticusKeeley24


def truncation_models(model_name):
    """
    Loads and returns methods to set the tidal truncation radius (or equivalent parameter) of halos
    """
    kwargs_model = {}
    if model_name == 'TRUNCATION_R50':
        kwargs_model['LOS_truncation_factor'] = 50
        # truncates NFW halos at r50 (for field halos, this is comparable to the splashback radius)
        return TruncationRN, kwargs_model
    elif model_name == 'TRUNCATION_RN':
        return TruncationRN, kwargs_model
    elif model_name == 'TRUNCATION_ROCHE':
        return TruncationRoche, kwargs_model
    elif model_name == 'TRUNCATION_ROCHE_GILMAN2020':
        kwargs_model['RocheNorm'] = 1.4
        kwargs_model['m_power'] = 1./3
        kwargs_model['RocheNu'] = 2./3
        return TruncationRoche, kwargs_model
    elif model_name == 'ADIABATIC_TIDES':
        return AdiabaticTidesTruncation, kwargs_model
    elif model_name == 'SPLASHBACK':
        return TruncationSplashBack, kwargs_model
    elif model_name == 'TRUNCATION_MEAN_DENSITY':
        return TruncateMeanDensity, kwargs_model
    elif model_name == 'TRUNCATION_GALACTICUS':
        return TruncationGalacticus, kwargs_model
    elif model_name == 'TRUNCATION_GALACTICUS_KEELEY24':
        return TruncationGalacticusKeeley24, kwargs_model
    elif model_name == 'CONSTANT':
        return ConstantTruncationArcsec, kwargs_model
    else:
        raise Exception('model '+str(model_name)+' not recognized')




import numpy.testing as npt
from pyHalo.Halos.lens_cosmo import LensCosmo
from pyHalo.Halos.HaloModels.powerlaw import PowerLawSubhalo, PowerLawFieldHalo
from pyHalo.Halos.HaloModels.NFW import NFWFieldHalo
from pyHalo.Halos.concentration import ConcentrationDiemerJoyce
from lenstronomy.LensModel.Profiles.splcore import SPLCORE
from lenstronomy.LensModel.Profiles.nfw import NFW
import pytest
from astropy.cosmology import FlatLambdaCDM
from pyHalo.Cosmology.cosmology import Cosmology

class TestSPLCORE(object):

    def setup_method(self):

        astropy = FlatLambdaCDM(70.0, 0.3)
        cosmo = Cosmology(astropy_instance=astropy)
        self.zhalo = 0.5
        self.zsource = 2.0
        self.lens_cosmo = LensCosmo(self.zhalo, self.zsource, cosmo)
        self.nfw_profile_lenstronomy = NFW()
        self.splcore = SPLCORE()
        self.truncation_class = None
        self.concentration_class = ConcentrationDiemerJoyce(self.lens_cosmo, scatter=False)

    def test_lenstronomy_params(self):

        m = 10 ** 8
        x = 0.5
        y = 1.0
        r3d = 100
        is_subhalo = False
        gamma = 2.7
        x_match = 2.9
        unique_tag = 1.0
        kwargs_profile = {'log_slope_halo': gamma, 'x_match': x_match, 'x_core_halo': 0.075,
                          'evaluate_mc_at_zlens': True}
        splcore = PowerLawFieldHalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, kwargs_profile,
                                   self.truncation_class, self.concentration_class, unique_tag)
        nfw = NFWFieldHalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, kwargs_profile,
                                   self.truncation_class, self.concentration_class, unique_tag)
        kwargs_nfw_profile = nfw.lenstronomy_params[0][0]
        kwargs_splcore_profile = splcore.lenstronomy_params[0][0]
        npt.assert_almost_equal(kwargs_splcore_profile['r_core'],
                                kwargs_nfw_profile['Rs'] * kwargs_profile['x_core_halo'], 5)
        id = splcore.lenstronomy_ID
        npt.assert_string_equal('SPL_CORE', id[0])

    def test_enclosed_mass(self):

        m = 10 ** 8
        x = 0.5
        y = 1.0
        r3d = 100
        is_subhalo = False
        gamma = 2.8
        x_match = 3.5
        unique_tag = 1.0
        kwargs_profile = {'log_slope_halo': gamma, 'x_match': x_match, 'x_core_halo': 0.075,
                          'evaluate_mc_at_zlens': True}
        splcore = PowerLawFieldHalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, kwargs_profile,
                                    self.truncation_class, self.concentration_class, unique_tag)
        nfw = NFWFieldHalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, kwargs_profile,
                                   self.truncation_class, self.concentration_class, unique_tag)

        kwargs_splcore = splcore.lenstronomy_params[0][0]
        kwargs_nfw = nfw.lenstronomy_params[0][0]
        rmatch = x_match * kwargs_nfw['Rs']
        m3d_nfw = self.nfw_profile_lenstronomy.mass_3d_lens(rmatch, kwargs_nfw['Rs'],
                                                            kwargs_nfw['alpha_Rs'])
        m3d = self.splcore.mass_3d_lens(rmatch, kwargs_splcore['sigma0'],
                                                         kwargs_splcore['r_core'], kwargs_splcore['gamma'])
        npt.assert_almost_equal(m3d_nfw/m3d, 1.0, 4)

    def test_concentration_redshift_eval(self):
        m = 10 ** 8
        x = 0.5
        y = 1.0
        r3d = 100
        gamma = 2.6
        x_match = 2.5
        unique_tag = 1.0
        kwargs_profile = {'log_slope_halo': gamma, 'x_match': x_match, 'x_core_halo': 0.075,
                          'evaluate_mc_at_zlens': False}
        splcore = PowerLawFieldHalo(m, x, y, r3d, self.zhalo, False, self.lens_cosmo, kwargs_profile,
                                   self.truncation_class, self.concentration_class, unique_tag)
        splcore_subhalo = PowerLawSubhalo(m, x, y, r3d, self.zhalo, True, self.lens_cosmo, kwargs_profile,
                                   self.truncation_class, self.concentration_class, unique_tag)
        c = splcore.profile_args[0]
        c_sub = splcore_subhalo.profile_args[0]
        npt.assert_equal(False, c==c_sub)

if __name__ == '__main__':
    pytest.main()



from pyHalo.Rendering.halo_population import HaloPopulation
from pyHalo.Rendering.MassFunctions.mass_function_base import CDMPowerLaw
from pyHalo.Rendering.MassFunctions.density_peaks import ShethTormen
from pyHalo.Rendering.SpatialDistributions.uniform import LensConeUniform
from pyHalo.Halos.lens_cosmo import LensCosmo
from pyHalo.Cosmology.cosmology import Cosmology
from pyHalo.Cosmology.geometry import Geometry
from pyHalo.Rendering.SpatialDistributions.nfw import ProjectedNFW
from pyHalo.utilities import generate_lens_plane_redshifts
import numpy.testing as npt
import pytest

class TestPopulation(object):

    def setup_method(self):

        cosmo = Cosmology()
        lens_cosmo = LensCosmo(0.5, 1.5, cosmo)
        geometry = Geometry(cosmo, 0.5, 1.5, 6.0, 'DOUBLE_CONE')
        model_list = ['SUBHALOS', 'LINE_OF_SIGHT', 'LINE_OF_SIGHT_NOSHEET', 'TWO_HALO']
        mass_function_class_list = [CDMPowerLaw, ShethTormen, ShethTormen, ShethTormen]
        kwargs_subhalos = {'log_mlow': 6.0, 'log_mhigh': 9.0, 'm_pivot': 1e8, 'power_law_index': -1.9,
                                'delta_power_law_index': 0.054, 'draw_poisson': False, 'log_m_host': 13.5,
                                'sigma_sub': 0.05,
                           'host_scaling_factor': 0.88, 'redshift_scaling_factor': 1.7}
        kwargs_los = {'log_mlow': 6.0, 'log_mhigh': 10.0, 'm_pivot': 1e8,
                                'LOS_normalization': 1.0, 'delta_power_law_index': 0.0}
        kwargs_two_halo = {'log_mlow': 6.0, 'log_mhigh': 10.0, 'm_pivot': 1e8,
                                'LOS_normalization': 1.0, 'delta_power_law_index': 0.0, 'log_m_host': 13.0}
        kwargs_mass_function_list = [kwargs_subhalos, kwargs_los, kwargs_los, kwargs_two_halo]
        spatial_distribution_class_list = [ProjectedNFW, LensConeUniform, LensConeUniform, LensConeUniform]
        kwargs_subhalos_spatial = {'m_host': 10**13.0, 'zlens': 0.5, 'rmax2d_arcsec': 3.0, 'r_core_units_rs': 0.2,
                                   'lens_cosmo': lens_cosmo}
        kwargs_los_spatial = {'cone_opening_angle': 6.0, 'geometry': geometry}
        lens_plane_redshift_list, redshift_spacings = generate_lens_plane_redshifts(0.5, 1.5)
        kwargs_spatial_distribution = [kwargs_subhalos_spatial, kwargs_los_spatial, kwargs_los_spatial, kwargs_los_spatial]
        self.model = HaloPopulation(model_list, mass_function_class_list, kwargs_mass_function_list,
                 spatial_distribution_class_list, kwargs_spatial_distribution,
                 lens_cosmo, geometry,
                 lens_plane_redshift_list, redshift_spacings)

    def test_render(self):

        masses, x, y, r3d, redshifts, is_subhalo_flag = self.model.render()
        npt.assert_equal(len(masses), len(x))
        npt.assert_equal(len(masses), len(y))
        npt.assert_equal(len(masses), len(r3d))
        npt.assert_equal(len(masses), len(redshifts))
        npt.assert_equal(len(masses), len(is_subhalo_flag))

if __name__ == '__main__':
   pytest.main()



import numpy.testing as npt
import numpy as np
from pyHalo.Halos.HaloModels.gaussian import Gaussian
from pyHalo.Halos.lens_cosmo import LensCosmo
from pyHalo.Cosmology.cosmology import Cosmology
from astropy.cosmology import FlatLambdaCDM
import pytest


class TestGaussianHalo(object):

    def setup_method(self):

        mass = 10 ** 8.
        x = 0.5
        y = 1.
        r3d = np.sqrt(1 + 0.5 ** 2 + 70 ** 2)
        astropy = FlatLambdaCDM(70.0, 0.3)
        cosmo = Cosmology(astropy)
        z = 0.5
        self.z = z
        lens_cosmo = LensCosmo(z, 2., cosmo)
        profile_args = {'amp':1,'sigma':1,'center_x':1.0,'center_y':1.0}
        sub_flag = False
        self.halo = Gaussian(mass, x, y, r3d, z,
                               sub_flag, lens_cosmo,
                               profile_args, None, None, unique_tag=np.random.rand())

    def test_lenstronomy_params(self):

        kwargs, _ = self.halo.lenstronomy_params
        for param in kwargs[0].keys():
            npt.assert_equal(kwargs[0][param], 1.0)

    def test_lenstronomy_ID(self):

        id = self.halo.lenstronomy_ID
        npt.assert_string_equal(id[0], 'GAUSSIAN')

    def test_redshift_eval(self):

        z_halo = self.halo.z_eval
        npt.assert_equal(z_halo, self.z)

if __name__ == '__main__':
   pytest.main()





from pyHalo.Halos.halo_base import Halo
import numpy as np
from lenstronomy.LensModel.Profiles.pseudo_jaffe import PseudoJaffe

class PJaffeSubhalo(Halo):
    """
    Class that defines a halo modeled as a Psuedo-Jaffe profile.

    The profile is normalized such that the Psuedo-Jaffe profile has the same mass within r200 as an NFW profile with mass
    M200. The scale radius of the Psuedo-Jafee profile is the same as the NFW profile

    """
    def __init__(self, mass, x, y, r3d, z,
                 sub_flag, lens_cosmo_instance, args, truncation_class, concentration_class, unique_tag):
        """
        See documentation in base class (Halos/halo_base.py)

        This profile is defined to have the same total mass as an NFW profile; rs = rs_pjaffe
        """
        self._lens_cosmo = lens_cosmo_instance
        self._prof = PseudoJaffe()
        self._concentration_class = concentration_class
        self._truncation_class = truncation_class
        mdef = 'PJAFFE'
        super(PJaffeSubhalo, self).__init__(mass, x, y, r3d, mdef, z, sub_flag,
                                            lens_cosmo_instance, args, unique_tag)

    @property
    def params_physical(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_params_physical'):

            concentration = self.profile_args
            rhos, rs, r200 = self._lens_cosmo.NFW_params_physical(self.mass, concentration, self.z)
            _, rs_kpc, r200_kpc = self._lens_cosmo.NFW_params_physical(self.mass, concentration, self.z)
            ra_kpc = 0.01 * rs_kpc

            rho = self._rho(self.mass, rs_kpc, ra_kpc, r200)

            self._params_physical = {'rhos': rhos, 'rs': rs, 'r200': r200, 'rho': rho}

        return self._params_physical

    @property
    def lenstronomy_params(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_lenstronomy_args'):

            kpc_to_arcsec = 1 / self._lens_cosmo.cosmo.kpc_proper_per_asec(self.z)
            (concentration) = self.profile_args
            rhos_kpc, rs_kpc, r200_kpc = self._lens_cosmo.NFW_params_physical(self.mass, concentration, self.z)
            ra_kpc = 0.01 * rs_kpc
            ra_arcsec = ra_kpc * kpc_to_arcsec
            rs_arcsec = rs_kpc * kpc_to_arcsec

            rmatch = r200_kpc
            rmatch_arcsec = rmatch * kpc_to_arcsec

            xmatch = rmatch/rs_kpc
            fx = np.log(1+xmatch) - xmatch / (1 + xmatch)
            m_nfw = 4 * np.pi * rs_kpc ** 3 * rhos_kpc * fx

            rho = self._rho(m_nfw, rs_arcsec, ra_arcsec, rmatch_arcsec)
            sigma0 = self._prof.rho2sigma(rho, ra_arcsec, rs_arcsec)
            sigma_crit_kpc = self._lens_cosmo.get_sigma_crit_lensing(self.z, self._lens_cosmo.z_source) * 0.001 ** 2
            sigma0 *= (sigma_crit_kpc/kpc_to_arcsec**2) ** -1

            self._lenstronomy_args = [{'center_x': self.x,
                                      'center_y': self.y,
                                      'Ra': ra_arcsec,
                                      'Rs': rs_arcsec,
                                      'sigma0': sigma0}]

        return self._lenstronomy_args, None

    @property
    def z_eval(self):
        """
        Returns the redshift at which to evalate the concentration-mass relation
        """
        if not hasattr(self, '_zeval'):

            if 'evaluate_mc_at_zlens' in self._args.keys() and self._args['evaluate_mc_at_zlens']:
                self._zeval = self.z
            else:
                self._zeval = self.z_infall

        return self._zeval

    def _rho(self, m, rs, ra, r_match):

        """
        returns the central density of a PJaffe halo such that the resulting halo has the mass m within r_match
        :param m:
        :param ra:
        :param r_match:
        :return:
        """
        f = (ra * np.arctan(r_match / ra) - rs * np.arctan(r_match / rs)) / (ra ** 2 - rs ** 2)
        rho = m / (4 * np.pi * ra ** 2 * rs ** 2 * f)
        return rho

    @property
    def lenstronomy_ID(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        return ['PJAFFE']

    @property
    def profile_args(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_profile_args'):

            concentration = self._concentration_class.nfw_concentration(self.mass, self.z_eval)
            self._profile_args = (concentration)

        return self._profile_args

class PJaffeFieldhalo(PJaffeSubhalo):
    """
    Defines a field halo modeled as a Psuedo-Jaffe profile
    """
    @property
    def profile_args(self):

        if not hasattr(self, '_profile_args'):

            concentration = self._concentration_class.nfw_concentration(self.mass, self.z)
            self._profile_args = (concentration)

        return self._profile_args







import warnings
import numpy as np
from scipy.special import hyp2f1
from scipy.integrate import quad
from scipy.interpolate import interp1d


# Implement the transfer function of subhalo density profile.

def _r_te_Du_2024(x, alpha=1.0, beta=3.0, gamma=1.0, delta=2.0):
    """
    Return the effective tidal (truncation) radius of a subhalo given the bound mass fraction.
    :param x: bound mass fraction
    :param alpha: alpha parameter in the generalized NFW profile
    :param beta: beta parameter in the generalized NFW profile
    :param gamma: gamma parameter in the generalized NFW profile
    :param delta: power index of the density truncation
    :return: the effective tidal (truncation) radius in units of infall virial radius
    """
    error_status = 0

    if (delta == 2.0):
        if (alpha == 1.0):
            if (beta == 3.0):
                if (gamma == 0.0):
                    A = -9.98429803e-01
                    B = 5.50290877e-04
                    C = 1.23471540e+00
                elif (gamma == 0.5):
                    A = -0.19210388
                    B = 0.42676044
                    C = 1.46250084
                elif (gamma == 1.0):
                    A = 0.68492777
                    B = 0.66438857
                    C = 2.07766512
                elif (gamma == 1.5):
                    A = 0.9839032
                    B = 0.76882264
                    C = 2.51841765
                else:
                    error_status = 1
            elif (beta == 4.0):
                if (gamma == 0.0):
                    A = 39.42152817
                    B = 0.58827911
                    C = 4.05363252
                elif (gamma == 0.5):
                    A = 2.06550731
                    B = 0.38713431
                    C = 3.63592324
                elif (gamma == 1.0):
                    A = 13.98955016
                    B = 0.72217703
                    C = 5.00880454
                elif (gamma == 1.5):
                    A = 83.39811168
                    B = 1.21616656
                    C = 6.62016776
                else:
                    error_status = 1
            else:
                error_status = 1
        elif (alpha == 2.0):
            if (beta == 3.0):
                if (gamma == 0.0):
                    A = -1.02415147e+00
                    B = 1.64345425e-03
                    C = 2.41297704e+00
                else:
                    error_status = 1
            elif (beta == 4.0):
                if (gamma == 0.0):
                    A = 3.27271231e-06
                    B = -1.94602678e+00
                    C = 5.84716369e+00
                else:
                    error_status = 1
                # throw a warning message if the x is too small, where the
                # fitting function might not work well.
                if (np.min(x) < 0.04):
                    warnings.warn(
                        'for [alpha=2, beta=4, gamma=0], the fitting function might not work reliably when the bound mass fraction is below 0.04')
            else:
                error_status = 1
        else:
            error_status = 1
    elif (delta == 3.0):
        if (alpha == 1.0):
            if (beta == 3.0):
                if (gamma == 0.0):
                    A = -0.93094144
                    B = 0.04703346
                    C = 0.76842857
                elif (gamma == 0.5):
                    A = -0.2290658
                    B = 0.41234666
                    C = 1.39936438
                elif (gamma == 1.0):
                    A = 0.90933265
                    B = 0.63676731
                    C = 2.18508243
                elif (gamma == 1.5):
                    A = 0.83533898
                    B = 0.73399035
                    C = 2.43215333
                else:
                    error_status = 1
            elif (beta == 4.0):
                if (gamma == 0.0):
                    A = 41.3266337
                    B = 0.60818104
                    C = 4.07045865
                elif (gamma == 0.5):
                    A = 0.42123475
                    B = -0.38164340
                    C = 3.62931001
                elif (gamma == 1.0):
                    A = 15.91886121
                    B = 0.71939954
                    C = 4.98186081
                elif (gamma == 1.5):
                    A = 104.0534991
                    B = 1.24696989
                    C = 6.62957129
                else:
                    error_status = 1
            else:
                error_status = 1
        elif (alpha == 2.0):
            if (beta == 3.0):
                if (gamma == 0.0):
                    A = -9.64566747e-01
                    B = -2.37053865e-03
                    C = 2.51094575e+00
                else:
                    error_status = 1
            elif (beta == 4.0):
                if (gamma == 0.0):
                    A = 5.79321402e-07
                    B = -1.62255246e+00
                    C = 5.47069835e+00
                else:
                    error_status = 1
                # throw a warning message if the x is too small, where the
                # fitting function might not work well.
                if (np.min(x) < 0.04):
                    warnings.warn(
                        'for [alpha=2, beta=4, gamma=0], the fitting function might not work reliably when the bound mass fraction is below 0.04')
            else:
                error_status = 1
        else:
            error_status = 1
    else:
        raise Exception('[delta] value not supported yet, set [delta] = 2 or 3')

    if (error_status):
        raise Exception('unsupported combination of [alpha, beta, gamma]')

    return (1.0 + A) * x ** B / (1.0 + A * x ** (2.0 * B)) / np.exp(C * (1.0 - x))


###############################################################################

def _f_t_Du_2024(x, alpha=1.0, beta=3.0, gamma=1.0, delta=2.0):
    """
    Return the density normalization of a subhalo given the bound mass fraction.
    :param x: bound mass fraction
    :param alpha: alpha parameter in the generalized NFW profile
    :param beta: beta parameter in the generalized NFW profile
    :param gamma: gamma parameter in the generalized NFW profile
    :param delta: power index of the density truncation
    :return: the density normalization
    """
    error_status = 0

    if (delta == 2.0):
        if (alpha == 1.0):
            if (beta == 3.0):
                if (gamma == 0.0):
                    D = 1.48242247
                    E = 0.63639230
                elif (gamma == 0.5):
                    D = 1.17122717
                    E = 0.45000174
                elif (gamma == 1.0):
                    D = 0.75826635
                    E = 0.23376409
                elif (gamma == 1.5):
                    D = -9.95103536e-01
                    E = 1.31454694e-04
                else:
                    error_status = 1
            elif (beta == 4.0):
                if (gamma == 0.0):
                    D = 1.18519061
                    E = 0.73729078
                elif (gamma == 0.5):
                    D = 0.75235790
                    E = 0.47771706
                elif (gamma == 1.0):
                    D = 0.36838147
                    E = 0.22307797
                elif (gamma == 1.5):
                    D = 0.28500656
                    E = 0.10528364
                else:
                    error_status = 1
            else:
                error_status = 1
        elif (alpha == 2.0):
            if (beta == 3.0):
                if (gamma == 0.0):
                    D = 1.37068893
                    E = 0.83149579
                else:
                    error_status = 1
            elif (beta == 4.0):
                if (gamma == 0.0):
                    D = 0.83836120
                    E = 1.23071929
                else:
                    error_status = 1
            else:
                error_status = 1
        else:
            error_status = 1
    elif (delta == 3.0):
        if (alpha == 1.0):
            if (beta == 3.0):
                if (gamma == 0.0):
                    D = 1.40209850
                    E = 0.63252747
                elif (gamma == 0.5):
                    D = 1.08663071
                    E = 0.45228951
                elif (gamma == 1.0):
                    D = 1.43637351
                    E = -0.24907289
                elif (gamma == 1.5):
                    D = 0.08092773
                    E = 0.08490512
                else:
                    error_status = 1
            elif (beta == 4.0):
                if (gamma == 0.0):
                    D = 1.08789526
                    E = 0.72799030
                elif (gamma == 0.5):
                    D = 0.68794863
                    E = 0.48619566
                elif (gamma == 1.0):
                    D = 0.33591171
                    E = 0.25084764
                elif (gamma == 1.5):
                    D = 0.15135479
                    E = 0.12838997
                else:
                    error_status = 1
            else:
                error_status = 1
        elif (alpha == 2.0):
            if (beta == 3.0):
                if (gamma == 0.0):
                    D = 1.31446961
                    E = 0.82402378
                else:
                    error_status = 1
            elif (beta == 4.0):
                if (gamma == 0.0):
                    D = 0.77798730
                    E = 1.21267992
                else:
                    error_status = 1
            else:
                error_status = 1
        else:
            error_status = 1
    else:
        raise Exception('[delta] value not supported yet, set [delta] = 2 or 3')

    if (error_status):
        raise Exception('unsupported combination of [alpha, beta, gamma]')

    # For x > 1, set f_t = 1.
    f_t = np.where(x <= 1.0, (1.0 + D) * x ** E / (1.0 + D * x ** (2.0 * E)), 1.0)

    # f_t should not be larger than 1.
    f_t = np.where(f_t <= 1.0, f_t, 1.0)

    return f_t


###############################################################################
#
# Note that the fittting functions above are derived in terms of the bound mass fraction
#
#   x = M_{bound} / M_{bound,0},
#
# and a fixed halo concentration of c_vir  = 20.6.
# For a different mass definition or halo concentration, the tracks of r_te(x) and f_t(x)
# might be different. So we redefine r_te and f_t in terms of the mass ratio
#
#   y = M_{bound} / M_{mx,0},
#
# where M_{mx,0} is the enclosed mass within the infall Rmax (the radius where the circular
# velocity reaches its maximum).
#
def _M_enclosed_generalized_NFW_dimensionless(x, alpha=1.0, beta=3.0, gamma=1.0):
    """
    Return the dimensionless enclosed mass within radius x = r / r_s for the generalized NFW profile.
    :param x: dimensionless radius
    :param alpha: alpha parameter in the generalized NFW profile
    :param beta: beta parameter in the generalized NFW profile
    :param gamma: gamma parameter in the generalized NFW profile
    :return: the dimensionless enclosed mass
    """
    status = 1

    # For special cases, use simple analytic expressions if available.
    if (alpha == 1.0):
        if (beta == 3.0):
            if (gamma == 0.0):
                mass = 4.0 \
                       * np.pi \
                       * ( \
                               +np.log(+1.0 + x) \
                               - x \
                               * (+2.0 + 3.0 * x) \
                               / 2.0 \
                               / (+1.0 + x) ** 2 \
                           )
            elif (gamma == 0.5):
                mass = +( \
                        -8.0 \
                        * np.pi \
                        * np.sqrt(x) \
                        * (3.0 + 4.0 * x) \
                    ) \
                       / ( \
                               +3.0 \
                               * (1.0 + x) ** 1.5 \
                           ) \
                       + 8.0 \
                       * np.pi \
                       * np.arcsinh(np.sqrt(x))
            elif (gamma == 1.0):
                mass = +4.0 \
                       * np.pi \
                       * ( \
                               +np.log(+1.0 + x) \
                               - x \
                               / (+1.0 + x) \
                           )
            elif (gamma == 1.5):
                mass = +8.0 \
                       * np.pi \
                       * ( \
                               -           np.sqrt(x / (1.0 + x)) \
                               + np.arcsinh(np.sqrt(x)) \
                           )
            else:
                status = 0
        elif (beta == 4.0):
            mass = 4.0 * np.pi / (3.0 - gamma) * (x / (1.0 + x)) ** (3.0 - gamma)
        else:
            status = 0
    elif (alpha == 2.0):
        if (beta == 3.0):
            if (gamma == 0.0):
                mass = -4.0 \
                       * np.pi \
                       * ( \
                               +x \
                               / np.sqrt(1.0 + x ** 2) \
                               + np.log(np.sqrt(1.0 + x ** 2) - x) \
                           )
            elif (gamma == 1.0):
                mass = 2.0 * np.pi * np.log(1.0 + x ** 2)
            else:
                status = 0
        elif (beta == 4.0):
            if (gamma == 0.0):
                mass = +2.0 \
                       * np.pi \
                       * ( \
                               +np.arctan(x) \
                               - x / (1.0 + x ** 2)
                       )
            elif (gamma == 1.0):
                mass = 4.0 * np.pi * (1.0 - 1.0 / np.sqrt(1.0 + x ** 2))
            else:
                status = 0
        else:
            status = 0
    else:
        status = 0

    if (status == 0):
        # Use the general expression.
        mass = +4.0 \
               * np.pi \
               * x ** (3.0 - gamma) \
               * hyp2f1((3.0 - gamma) / alpha, (beta - gamma) / alpha, 1.0 + (3.0 - gamma) / alpha, -x ** alpha) \
               / (3.0 - gamma)

    return mass


def _R_max_generalized_NFW_dimensionless(alpha=1.0, beta=3.0, gamma=1.0):
    """
    Return the dimensionless Rmax for the generalized NFW profile.
    :param x: dimensionless radius
    :param alpha: alpha parameter in the generalized NFW profile
    :param beta: beta parameter in the generalized NFW profile
    :param gamma: gamma parameter in the generalized NFW profile
    :return: the dimensionless Rmax
    """
    status = 1

    # Use expressions for special cases if available.
    if (alpha == 1.0):
        if (beta == 3.0):
            if (gamma < 2.0):
                if (gamma == 0.0):
                    Rmax = 4.424700646872270
                elif (gamma == 0.5):
                    Rmax = 3.289276561384110
                elif (gamma == 1.0):
                    Rmax = 2.162581587064612
                elif (gamma == 1.5):
                    Rmax = 1.054966571869124
                else:
                    status = 0
            else:
                raise Exception('Rmax does not exist')
        elif (beta == 4.0):
            if (gamma < 2.0):
                Rmax = 2.0 - gamma
            else:
                raise Exception('Rmax does not exist')
        else:
            status = 0
    elif (alpha == 2.0):
        if (beta == 3.0):
            if (gamma < 2.0):
                if (gamma == 0.0):
                    Rmx = 2.919847688299723
                else:
                    status = 0
            else:
                raise Exception('Rmax does not exist')
        elif (beta == 4.0):
            if (gamma < 2.0):
                if (gamma == 0.0):
                    Rmx = 1.825255642519694
                else:
                    status = 0
            else:
                raise Exception('Rmax does not exist')
        else:
            status = 0
    else:
        status = 0

    if (status == 0):
        # Use numerical solver.
        def root(x):
            return (1.0 + x ** alpha) ** ((gamma - beta) / alpha) \
                   - hyp2f1((3.0 - gamma) / alpha, (beta - gamma) / alpha, 1.0 + (3.0 - gamma) / alpha, -x ** alpha) \
                   / (3.0 - gamma)

        Rmax = newton(root, 1.0)

    return Rmax


def _M_total_generalized_NFW_Truncated_dimensionless(tau, alpha=1.0, beta=3.0, gamma=1.0, delta=2.0):
    """
    Return the dimensionless total mass of a subhalo with truncated generalized NFW profile:
    rho(x) = rho_org(x) / (1.0 + (x / tau)**2),
    where rho_org(x) is the original density profile (without truncation), x = r / r_s, and tau = r_te / r_s.
    :param tau: dimensionless truncation radius
    :param alpha: alpha parameter in the generalized NFW profile
    :param beta: beta parameter in the generalized NFW profile
    :param gamma: gamma parameter in the generalized NFW profile
    :param delta: power index of the density truncation
    :return: the dimensionless enclosed mass
    """
    status = 1

    if (delta == 2.0):
        if (alpha == 1.0):
            if (beta == 3.0):
                if (gamma == 0.0):
                    mass = 4.0 * np.pi \
                           * tau ** 2 / (tau ** 2 + 1.0) ** 3 \
                           * ( \
                                   0.5 * (-1.0 + (np.pi - tau) * tau) * (3.0 * tau ** 2 - 1.0) \
                                   + tau ** 2 * (tau ** 2 - 3.0) * np.log(tau) \
                               )
                elif (gamma == 0.5):
                    atan_tau = np.arctan(tau)
                    mass = 4.0 * np.pi \
                           * tau ** 2 / (tau ** 2 + 1.0) ** (9.0 / 4.0) \
                           / 6.0 \
                           * ( \
                                   3.0 * np.sqrt(2.0) * np.pi * np.sqrt(tau) * (tau ** 2 + 1.0) \
                                   * (np.sin(2.5 * atan_tau) - np.cos(2.5 * atan_tau)) \
                                   + 2.0 * np.sqrt(2.0) \
                                   * (np.sqrt(1.0 + np.sqrt(2.0 * tau) + tau) - np.sqrt(1.0 - np.sqrt(2.0 * tau) + tau)) \
                                   * np.sqrt(tau * (tau ** 2 + 1.0) * (+1.0 - tau + np.sqrt(tau ** 2 + 1.0))) \
                                   * np.cos(0.25 * (np.pi + 2.0 * atan_tau)) \
                                   - 2.0 * np.sqrt(2.0) \
                                   * (np.sqrt(1.0 + np.sqrt(2.0 * tau) + tau) + np.sqrt(1.0 - np.sqrt(2.0 * tau) + tau)) \
                                   * np.sqrt(tau * (tau ** 2 + 1.0) * (-1.0 + tau + np.sqrt(tau ** 2 + 1.0))) \
                                   * np.sin(0.25 * (np.pi + 2.0 * atan_tau)) \
                                   + ( \
                                           +( \
                                                   +2.0 * np.sqrt(2.0) + 4.0 * np.sqrt(tau) + 3.0 * np.sqrt(2.0) * tau \
                                                   - 2.0 * np.sqrt(2.0) * tau ** 2 - 4.0 * tau ** 2.5 - np.sqrt(
                                                   2.0) * tau ** 3 \
                                               ) * np.sqrt(1.0 - np.sqrt(2.0 * tau) + tau) \
                                           + ( \
                                                   +2.0 * np.sqrt(2.0) - 4.0 * np.sqrt(tau) + 3.0 * np.sqrt(2.0) * tau \
                                                   - 2.0 * np.sqrt(2.0) * tau ** 2 + 4.0 * tau ** 2.5 - np.sqrt(
                                                   2.0) * tau ** 3 \
                                               ) * np.sqrt(1.0 + np.sqrt(2.0 * tau) + tau) \
                                       ) \
                                   * np.sqrt((-1.0 + tau + np.sqrt(tau ** 2 + 1.0)) / tau) \
                                   * np.cos(0.25 * (np.pi + 2.0 * atan_tau)) \
                                   + ( \
                                           +( \
                                                   +2.0 * np.sqrt(2.0) + 4.0 * np.sqrt(tau) + 3.0 * np.sqrt(2.0) * tau \
                                                   - 2.0 * np.sqrt(2.0) * tau ** 2 - 4.0 * tau ** 2.5 - np.sqrt(
                                                   2.0) * tau ** 3 \
                                               ) * np.sqrt(1.0 - np.sqrt(2.0 * tau) + tau) \
                                           - ( \
                                                   +2.0 * np.sqrt(2.0) - 4.0 * np.sqrt(tau) + 3.0 * np.sqrt(2.0) * tau \
                                                   - 2.0 * np.sqrt(2.0) * tau ** 2 + 4.0 * tau ** 2.5 - np.sqrt(
                                                   2.0) * tau ** 3 \
                                               ) * np.sqrt(1.0 + np.sqrt(2.0 * tau) + tau) \
                                       ) \
                                   * np.sqrt((+1.0 - tau + np.sqrt(tau ** 2 + 1.0)) / tau) \
                                   * np.sin(0.25 * (np.pi + 2.0 * atan_tau)) \
                                   - 12.0 \
                                   * ( \
                                           +(tau ** 2 - 1.0) * np.arcsin(
                                           np.sqrt(0.5 * (1.0 + tau - np.sqrt(tau ** 2 + 1.0)))) \
                                           + tau \
                                           * np.log(tau + np.sqrt(tau ** 2 + 1.0) + np.sqrt(
                                           2.0 * tau * (tau + np.sqrt(tau ** 2 + 1.0)))) \
                                       ) \
                                   * np.sqrt(tau) \
                                   * np.cos(0.25 * (np.pi + 2.0 * atan_tau)) \
                                   - 12.0 \
                                   * ( \
                                           +2.0 * tau * np.arcsin(np.sqrt(0.5 * (1.0 + tau - np.sqrt(tau ** 2 + 1.0)))) \
                                           - 0.5 * (tau ** 2 - 1.0) \
                                           * np.log(tau + np.sqrt(tau ** 2 + 1.0) + np.sqrt(
                                           2.0 * tau * (tau + np.sqrt(tau ** 2 + 1.0)))) \
                                       ) \
                                   * np.sqrt(tau) \
                                   * np.sin(0.25 * (np.pi + 2.0 * atan_tau)) \
                               )
                elif (gamma == 1.0):
                    mass = 4.0 * np.pi \
                           * tau ** 2 / (tau ** 2 + 1.0) ** 2 \
                           * ( \
                                   (tau ** 2 - 1.0) * np.log(tau) \
                                   + tau * np.pi \
                                   - (tau ** 2 + 1.0)
                           )
                elif (gamma == 1.5):
                    atan_tau = np.arctan(tau)
                    mass = 4.0 * np.pi \
                           * tau ** 1.5 / (tau ** 2 + 1.0) ** (5.0 / 4.0) \
                           / 2.0 \
                           * ( \
                                   +np.sqrt(2.0) * np.pi * np.sqrt(tau ** 2 + 1.0) \
                                   * (np.cos(1.5 * atan_tau) + np.sin(1.5 * atan_tau)) \
                                   - np.sqrt(2.0) \
                                   * ( \
                                           +np.sqrt(1.0 + np.sqrt(2.0 * tau) + tau) \
                                           + np.sqrt(1.0 - np.sqrt(2.0 * tau) + tau) \
                                       ) \
                                   * np.sqrt((tau ** 2 + 1.0) * (-1.0 + tau + np.sqrt(tau ** 2 + 1.0))) \
                                   * np.cos(0.25 * (np.pi + 2.0 * atan_tau)) \
                                   - np.sqrt(2.0) \
                                   * ( \
                                           +np.sqrt(1.0 + np.sqrt(2.0 * tau) + tau) \
                                           - np.sqrt(1.0 - np.sqrt(2.0 * tau) + tau) \
                                       ) \
                                   * np.sqrt((tau ** 2 + 1.0) * (+1.0 - tau + np.sqrt(tau ** 2 + 1.0))) \
                                   * np.sin(0.25 * (np.pi + 2.0 * atan_tau)) \
                                   - 2.0 \
                                   * ( \
                                           +2.0 * tau * np.arcsin(np.sqrt(0.5 * (1.0 + tau - np.sqrt(tau ** 2 + 1.0)))) \
                                           + np.log(tau + np.sqrt(tau ** 2 + 1.0) + np.sqrt(
                                           2.0 * tau * (tau + np.sqrt(tau ** 2 + 1.0)))) \
                                       ) \
                                   * np.cos(0.25 * (np.pi + 2.0 * atan_tau)) \
                                   - 2.0 \
                                   * ( \
                                           +2.0 * np.arcsin(np.sqrt(0.5 * (1.0 + tau - np.sqrt(tau ** 2 + 1.0)))) \
                                           - tau \
                                           * np.log(tau + np.sqrt(tau ** 2 + 1.0) + np.sqrt(
                                           2.0 * tau * (tau + np.sqrt(tau ** 2 + 1.0)))) \
                                       ) \
                                   * np.sin(0.25 * (np.pi + 2.0 * atan_tau)) \
                               )
                else:
                    status = 0
            else:
                status = 0
        elif (alpha == 2.0):
            if (beta == 4.0):
                if (gamma == 0.0):
                    mass = np.pi ** 2 * tau ** 2 / (tau + 1.0) ** 2
                else:
                    status = 0
            else:
                status = 0
        else:
            status = 0
    elif (delta == 3.0):
        if (alpha == 1.0 and beta == 3.0 and gamma == 1.0):
            mass = 4.0 * np.pi \
                   * tau ** 2 / (tau ** 3 - 1.0) ** 2 \
                   * ( \
                           -tau * (tau ** 3 - 1.0) \
                           + 2.0 * np.sqrt(3.0) * np.pi / 9.0 \
                           * (tau - 1.0) ** 2 \
                           * (2.0 * tau + 1.0) \
                           + tau \
                           * (tau ** 3 + 2.0) \
                           * np.log(tau) \
                       )
        else:
            status = 0
    else:
        status = 0

    if (status == 0):
        # Use numerical integration.
        def integral(x):
            return 4.0 * np.pi * x ** 2 \
                   / x ** gamma \
                   / (1.0 + x ** alpha) ** ((beta - gamma) / alpha) \
                   / (1.0 + (x / tau) ** delta)

        mass = quad(integral, 0.0, np.inf)

    return mass


# Mass table used to compute the radius enclosing a certain mass for truncated generalized NFW profile.
N_per_decade = 10
x_min = 1.0e-6
x_max = 1.0e+4
N_tab = int(np.log10(x_max / x_min) * N_per_decade)
x_tab = np.geomspace(x_min, x_max, N_tab)
M_tab = np.zeros(N_tab)

Mass_table_Initialized = False
alpha_used = None
beta_used = None
gamma_used = None
delta_used = None
M_min = np.inf
M_max = 0.0


def _Truncation_Radius(M, f_t, alpha=1.0, beta=3.0, gamma=1.0, delta=2.0):
    """
    Return the truncation radius of a subhalo given the bound mass and normalization of density profile.
    The density profile is assumed to be
    rho(x) = rho_org(x) * f_t / (1.0 + (x / tau)**2),
    where rho_org(x) is the original density profile (without truncation), x = r / r_s, and tau = r_te / r_s.
    :param M: dimensionless mass of the subhalo
    :param f_t: normalization of the density profile
    :param alpha: alpha parameter in the generalized NFW profile
    :param beta: beta parameter in the generalized NFW profile
    :param gamma: gamma parameter in the generalized NFW profile
    :param delta: power index of the density truncation
    :return: the dimensionless truncation radius
    """

    global N_per_decade, x_min, x_max, N_tab, x_tab, M_tab
    global Mass_table_Initialized, alpha_used, beta_used, gamma_used, delta_used, M_min, M_max

    M_unnormalized = M / f_t

    retabulate = True
    if (Mass_table_Initialized and alpha_used == alpha and beta_used == beta and gamma_used == gamma \
        and delta_used == delta and np.min(M_unnormalized) >= M_min and np.max(M_unnormalized) <= M_max):
        retabulate = False

    if (retabulate):
        M_min = _M_total_generalized_NFW_Truncated_dimensionless(x_min, alpha, beta, gamma, delta)
        M_max = _M_total_generalized_NFW_Truncated_dimensionless(x_max, alpha, beta, gamma, delta)
        while (np.min(M_unnormalized) < M_min):
            x_min = x_min / 2.0
            M_min = _M_total_generalized_NFW_Truncated_dimensionless(x_min, alpha, beta, gamma, delta)

        while (np.max(M_unnormalized) > M_max):
            x_max = x_max * 2.0
            M_max = _M_total_generalized_NFW_Truncated_dimensionless(x_max, alpha, beta, gamma, delta)

        N_tab = int(np.log10(x_max / x_min) * N_per_decade)
        x_tab = np.geomspace(x_min, x_max, N_tab)
        M_tab = _M_total_generalized_NFW_Truncated_dimensionless(x_tab, alpha, beta, gamma, delta)

        Mass_table_Initialized = True
        alpha_used = alpha
        beta_used = beta
        gamma_used = gamma
        delta_used = delta

    x_M_interp_loglog = interp1d(np.log(M_tab), np.log(x_tab), kind='cubic')

    return np.exp(x_M_interp_loglog(np.log(M_unnormalized)))


def Convert_to_reference_model(alpha=1.0, beta=3.0, gamma=1.0):
    """
    Compute the conversion factors from parameter y = M_{bound} / M_{mx,0} to parameter
    x = M_{bound} / M_{bound,0} in the reference model.
    :param alpha: alpha parameter in the generalized NFW profile
    :param beta: beta parameter in the generalized NFW profile
    :param gamma: gamma parameter in the generalized NFW profile
    :return: conversion factors [Mmx_ref / Mbound_ref_0, Rmax_ref / Rvir_ref]
    """
    error_status = 0

    if (alpha == 1.0):
        if (beta == 3.0):
            if (gamma == 0.0):
                Mbound_ref_0 = 1.1969934e9  # in units of Msun
            elif (gamma == 0.5):
                Mbound_ref_0 = 1.1793903e9  # in units of Msun
            elif (gamma == 1.0):
                Mbound_ref_0 = 1.1605002e9  # in units of Msun
            elif (gamma == 1.5):
                Mbound_ref_0 = 1.1395515e9  # in units of Msun
            else:
                error_status = 1
        elif (beta == 4.0):
            if (gamma == 0.0):
                Mbound_ref_0 = 1.0426536e9  # in units of Msun
            elif (gamma == 0.5):
                Mbound_ref_0 = 1.0353760e9  # in units of Msun
            elif (gamma == 1.0):
                Mbound_ref_0 = 1.0281669e9  # in units of Msun
            elif (gamma == 1.5):
                Mbound_ref_0 = 1.0210254e9  # in units of Msun
            else:
                error_status = 1
        else:
            error_status = 1
    elif (alpha == 2.0):
        if (beta == 3.0):
            if (gamma == 0.0):
                Mbound_ref_0 = 1.1343262e9  # in units of Msun
            else:
                error_status = 1
        elif (beta == 4.0):
            if (gamma == 0.0):
                Mbound_ref_0 = 1.0194281e9  # in units of Msun
            else:
                error_status = 1
        else:
            error_status = 1
    else:
        error_status = 1

    if (error_status):
        raise Exception('unsupported combination of [alpha, beta, gamma]')

    Mvir_ref = 1.0e9  # in units of Msun, note that Mbound_ref_0 is slightly larger than Mvir_ref
    Rvir_ref = 0.02632439488393568  # in units of Mpc
    c_ref = 20.587820037799197
    Rmax_dimensionless = _R_max_generalized_NFW_dimensionless(alpha, beta, gamma)
    Rmax_ref = Rmax_dimensionless / c_ref * Rvir_ref
    Mmx_ref = _M_enclosed_generalized_NFW_dimensionless(Rmax_dimensionless, alpha, beta, gamma) \
              / _M_enclosed_generalized_NFW_dimensionless(c_ref, alpha, beta, gamma) \
              * Mvir_ref

    return Mmx_ref / Mbound_ref_0, Rmax_ref / Rvir_ref


def compute_r_te_and_f_t(Mbound, MvirInfall, RvirInfall, cInfall, alpha=1.0, beta=3.0, gamma=1.0, delta=2.0):
    """
    Return the effective tidal (truncation) radius and the density normalization of a subhalo given the
    bound mass and infall halo properties.
    :param Mbound: bound mass
    :param MvirInfall: virial mass at infall
    :param RvirInfall: virial radius at infall
    :param cInfall: halo concentration at infall
    :param alpha: alpha parameter in the generalized NFW profile
    :param beta: beta parameter in the generalized NFW profile
    :param gamma: gamma parameter in the generalized NFW profile
    :param delta: power index of the density truncation
    :return: the effective tidal (truncation) radius
    """

    Rs = RvirInfall / cInfall
    Rmax_dimensionless = _R_max_generalized_NFW_dimensionless(alpha, beta, gamma)

    # Characteristic mass.
    M0 = MvirInfall / _M_enclosed_generalized_NFW_dimensionless(cInfall, alpha, beta, gamma)

    M_mx = M0 * _M_enclosed_generalized_NFW_dimensionless(Rmax_dimensionless, alpha, beta, gamma)

    y = Mbound / M_mx

    # Convert y to parameter x = MBoud / MvirInfall in the reference model.
    y_scale, R_scale = Convert_to_reference_model(alpha, beta, gamma)

    # r_te = r_te_Du_2024(y * y_scale, alpha, beta, gamma, delta) / R_scale * Rmax
    ## Make sure that the truncation radius is smaller than the virial radius at infall.
    # r_te = np.where(r_te < RvirInfall, r_te, RvirInfall)

    f_t = _f_t_Du_2024(y * y_scale, alpha, beta, gamma, delta)

    # Instead of directly using the fitting function for r_te, we first compute f_t
    # using the fitting function and then compute r_te assuming that the total mass
    # of the truncated halo equals to the bound mass. r_te computed in this way is
    # in good agreement with the fitting function, but is more reliable where the
    # fitting function is not calibrated to simulations.
    Mbound_dimensionless = Mbound / M0
    r_te = Rs * _Truncation_Radius(Mbound_dimensionless, f_t, alpha, beta, gamma)

    return r_te, f_t



import pytest
import numpy as np
from pyHalo.Halos.lens_cosmo import LensCosmo
from pyHalo.Cosmology.cosmology import Cosmology
import numpy.testing as npt
from pyHalo.Halos.concentration import ConcentrationLudlow, ConcentrationDiemerJoyce, \
    ConcentrationPeakHeight, ConcentrationWDMPolynomial, ConcentrationWDMHyperbolic, ConcentrationLudlowWDM
from astropy.cosmology import FlatLambdaCDM

class TestConcentration(object):

    def setup_method(self):

        astropy = FlatLambdaCDM(70.0, 0.3)
        cosmo = Cosmology(astropy_instance=astropy)
        self.astropy = cosmo.astropy

    def test_concentration_ludlow_wdm(self):

        scatter = False
        z = 0.0
        log_mc = 7.0
        slope = -1.0
        m = 10 ** 7.5

        concentration_model_cdm = ConcentrationLudlow(self.astropy, scatter)
        concentration_model_wdm = ConcentrationLudlowWDM(self.astropy, log_mc, slope, scatter)
        a_true, b_true = 0.473, 0.739
        a_model, b_model = concentration_model_wdm.suppression_fit(log_mc, slope, z)
        npt.assert_almost_equal(a_model, a_true, 2)
        npt.assert_almost_equal(b_model, b_true, 2)
        arg = (np.log10(m / 10 ** log_mc) - a_model) / (2 * b_model)
        suppression = 0.5 * (1.0 + np.tanh(arg))

        c_wdm = concentration_model_wdm.nfw_concentration(m, z)
        c_cdm = concentration_model_cdm.nfw_concentration(m, z)
        ratio = c_wdm / c_cdm
        npt.assert_almost_equal(ratio, 0.50968, 3)
        npt.assert_almost_equal(c_wdm, c_cdm * suppression, 3)

        z = 1.0
        log_mc = 6.0
        slope = -3.2
        m = 10 ** 6.0

        concentration_model_cdm = ConcentrationLudlow(self.astropy, scatter)
        concentration_model_wdm = ConcentrationLudlowWDM(self.astropy, log_mc, slope, scatter)
        a_true, b_true = 0.42797, 0.464066
        a_model, b_model = concentration_model_wdm.suppression_fit(log_mc, slope, z)
        npt.assert_almost_equal(a_model, a_true, 2)
        npt.assert_almost_equal(b_model, b_true, 2)
        arg = (np.log10(m / 10 ** log_mc) - a_model) / (2 * b_model)
        suppression = 0.5 * (1.0 + np.tanh(arg))

        c_wdm = concentration_model_wdm.nfw_concentration(m, z)
        c_cdm = concentration_model_cdm.nfw_concentration(m, z)
        ratio = c_wdm / c_cdm
        npt.assert_almost_equal(ratio, 0.28453, 3)
        npt.assert_almost_equal(c_wdm, c_cdm * suppression, 3)

        log10_mhm_eval, dlogT_dlogk_eval, z_eval = 9.0, -4.5, 4.5
        log10_mhm_eval, dlogT_dlogk_eval, z_eval = concentration_model_wdm._make_in_bounds(log10_mhm_eval, dlogT_dlogk_eval, z_eval)
        npt.assert_equal(log10_mhm_eval, 8.0)
        npt.assert_equal(dlogT_dlogk_eval, 4.0)
        npt.assert_equal(z_eval, 4.0)

        log10_mhm_eval, dlogT_dlogk_eval, z_eval = 5.6, -0.5, 4.0
        log10_mhm_eval, dlogT_dlogk_eval, z_eval = concentration_model_wdm._make_in_bounds(log10_mhm_eval,
                                                                                           dlogT_dlogk_eval, z_eval)
        npt.assert_equal(log10_mhm_eval, 6.0)
        npt.assert_equal(dlogT_dlogk_eval, 1.0)
        npt.assert_equal(z_eval, 4.0)

    def test_concentration_diemer_joyce(self):

            m = 10 ** 8
            z = 0.5
            c_true = 14.246936385951503
            scatter = False
            scatter_amplitude_dex = 0.2
            concentration_model = ConcentrationDiemerJoyce(self.astropy, scatter)
            c = concentration_model.nfw_concentration(m, z)
            npt.assert_almost_equal(c_true, c)

            scatter = True
            concentration_model = ConcentrationDiemerJoyce(self.astropy, scatter)
            c = concentration_model.nfw_concentration(m, z)
            npt.assert_equal(c != c_true, True)

            c = concentration_model.nfw_concentration(np.array([10**8]*10), z)
            npt.assert_equal(10, len(c))

    def test_concentration_ludlow(self):

        m = 10 ** 8
        z = 0.5
        c_true = 13.36631
        scatter = False
        concentration_model = ConcentrationLudlow(self.astropy, scatter)
        c = concentration_model.nfw_concentration(m, z)
        npt.assert_almost_equal(c_true, c, 3)

        scatter = True
        concentration_model = ConcentrationDiemerJoyce(self.astropy, scatter)
        c = concentration_model.nfw_concentration(m, z)
        npt.assert_equal(c != c_true, True)

        c = concentration_model.nfw_concentration(np.array([10 ** 8] * 10), z)
        npt.assert_equal(10, len(c))


    def test_concentration_peak_height(self):

        m = 10 ** 8
        concentration_model_1 = ConcentrationPeakHeight(self.astropy, 16.0, -0.2, 0.8, False)
        concentration_model_2 = ConcentrationPeakHeight(self.astropy, 16.0, -0.2, 4.0, False)
        concentration_model_3 = ConcentrationPeakHeight(self.astropy, 2.5 * 16.0, -0.2, 4.0, False)
        c1 = concentration_model_1.nfw_concentration(m, 0.5)
        c2 = concentration_model_2.nfw_concentration(m, 0.5)
        c3 = concentration_model_3.nfw_concentration(m, 0.5)
        npt.assert_equal(c1, c2)
        npt.assert_almost_equal(c1, c3 / 2.5, 6)

    def test_concentration_wdm_polynomial(self):

        m = 10 ** 8
        z = 0.5
        c0 = 21.49
        zeta = -0.5
        beta = 0.8
        kwargs_cdm = {'c0': c0, 'zeta': zeta, 'beta': beta, 'scatter': False}
        concentration_cdm = ConcentrationPeakHeight
        concentration_cdm_peak_height = concentration_cdm(self.astropy, **kwargs_cdm)
        log_mc = 8.2
        c_scale = 60.0
        c_power = -0.17
        c_power_inner = 1.5

        mc_suppression_redshift_evolution = False
        concentration_model = ConcentrationWDMPolynomial(self.astropy, concentration_cdm,
                                                         log_mc, c_scale, c_power, c_power_inner, mc_suppression_redshift_evolution,
                                                         scatter=False, kwargs_cdm=kwargs_cdm)

        c_wdm = concentration_model.nfw_concentration(m, z)
        c_cdm = concentration_cdm_peak_height.nfw_concentration(m, z)
        suppression = (1 + c_scale * (10**log_mc / m)**c_power_inner)**c_power
        npt.assert_almost_equal(c_wdm, c_cdm * suppression)

        mc_suppression_redshift_evolution = True
        concentration_model = ConcentrationWDMPolynomial(self.astropy, concentration_cdm,
                                                         log_mc, c_scale, c_power, c_power_inner,
                                                         mc_suppression_redshift_evolution,
                                                         scatter=False, kwargs_cdm=kwargs_cdm)

        c_wdm = concentration_model.nfw_concentration(m, z)
        c_cdm = concentration_cdm_peak_height.nfw_concentration(m, z)
        suppression *= (1 + z) ** (0.026 * z - 0.04)
        npt.assert_almost_equal(c_wdm, c_cdm * suppression)

    def test_concentration_wdm_hyperbolic(self):

        m = 10 ** 8
        z = 0.5
        kwargs_cdm = {'scatter': False}
        concentration_cdm = ConcentrationDiemerJoyce
        concentration_cdm_diemer_joyce = concentration_cdm(self.astropy, **kwargs_cdm)
        log_mc = 8.2
        a = 0.2
        b = 0.5
        concentration_model = ConcentrationWDMHyperbolic(self.astropy, concentration_cdm, log_mc, a, b,
                                                         scatter=False,kwargs_cdm=kwargs_cdm)
        c_wdm = concentration_model.nfw_concentration(m, z)
        c_cdm = concentration_cdm_diemer_joyce.nfw_concentration(m, z)
        suppression = 0.5 * (1 + np.tanh((np.log10(m/10**log_mc) - a) / (2 * b)))
        npt.assert_almost_equal(c_wdm, c_cdm * suppression)

if __name__ == '__main__':
    pytest.main()



import pytest
import numpy as np
from pyHalo.Cosmology.cosmology import Cosmology
from pyHalo.Cosmology.geometry import Geometry
import numpy.testing as npt
from pyHalo.Rendering.MassFunctions.delta_function import DeltaFunction

class TestDeltaFunction(object):

    def setup_method(self):

        z = 0.5
        delta_z = 0.02
        opening_angle = 6.0
        zlens = 0.5
        zsource = 1.5
        cosmo = Cosmology()
        astropy = cosmo.astropy
        geometry_class = Geometry(cosmo, zlens, zsource, opening_angle, 'DOUBLE_CONE')
        density_to_MsunperMpc = geometry_class.cosmo.density_to_MsunperMpc
        self._rho = astropy.Odm(0.) * astropy.critical_density(z).value * density_to_MsunperMpc
        self._volume = geometry_class.volume_element_comoving(z, delta_z)
        self._mass = 10 ** 7
        mass_fraction = 1.0
        self._kwargs_model = {'mass_fraction': mass_fraction, 'mass': self._mass,
                              'draw_poisson': False, 'LOS_normalization': 1.0}
        draw_poisson = False
        self.mass_function = DeltaFunction(self._mass, self._volume, mass_fraction*self._rho, draw_poisson)
        draw_poisson = True
        self.mass_function_poisson = DeltaFunction(self._mass, self._volume, mass_fraction*self._rho, draw_poisson)
        self._kwargs_from_redshift = {'z': z, 'delta_z': delta_z, 'geometry_class': geometry_class,
                                      'kwargs_model': self._kwargs_model}

    def test_mass_function(self):

        m = self.mass_function.draw()
        npt.assert_equal(m[0]==self._mass, True)
        n_theory = np.round(self._kwargs_model['mass_fraction']*self._rho * self._volume/self._mass)
        npt.assert_equal(True, len(m)==n_theory)
    #
    # def test_draw_poisson(self):
    #
    #     m = self.mass_function.draw()
    #     m_poisson = self.mass_function_poisson.draw()
    #     npt.assert_equal(True, len(m)!=len(m_poisson))

    def test_from_redshift(self):

        mfunc = DeltaFunction.from_redshift(**self._kwargs_from_redshift)
        m = mfunc.draw()
        npt.assert_equal(True, len(m)==len(self.mass_function.draw()))

if __name__ == '__main__':
   pytest.main()



import pytest
import numpy as np
from pyHalo.utilities import generate_lens_plane_redshifts
from pyHalo.Cosmology.cosmology import Cosmology
from pyHalo.Cosmology.geometry import Geometry
import numpy.testing as npt
from pyHalo.Halos.lens_cosmo import LensCosmo
from pyHalo.Rendering.SpatialDistributions.uniform import LensConeUniform
from pyHalo.Rendering.MassFunctions.density_peaks import ShethTormen
from pyHalo.Rendering.line_of_sight import LineOfSightNoSheet, LineOfSight

class TestLineOfSightNoSheet(object):

    def setup_method(self):

        mass_function_model = ShethTormen
        kwargs_mass_function = {'log_mlow': 6.0, 'log_mhigh': 10.0, 'm_pivot': 1e8,
                                'LOS_normalization': 1.0, 'delta_power_law_index': 0.0}
        zlens = 0.5
        zsource = 1.5
        cosmo = Cosmology()
        geometry = Geometry(cosmo, zlens, zsource, 6.0, 'DOUBLE_CONE')
        lens_cosmo = LensCosmo(zlens, zsource, cosmo)
        lens_plane_redshifts, delta_z_list = generate_lens_plane_redshifts(zlens, zsource)
        spatial_distribution_model = LensConeUniform(geometry.cone_opening_angle, geometry)
        self.model = LineOfSightNoSheet(mass_function_model, kwargs_mass_function, spatial_distribution_model,
                 geometry, lens_cosmo, lens_plane_redshifts, delta_z_list)
        self._is_subhalo = False

    def test_name(self):
        npt.assert_string_equal('LINE_OF_SIGHT_NO_SHEET', self.model.name)

    def test_render(self):

        masses, x, y, r3d, redshifts, subhalo_flag = self.model.render()
        npt.assert_equal(len(masses), len(x))
        npt.assert_equal(len(masses), len(y))
        npt.assert_equal(len(redshifts), len(masses))
        for flag in subhalo_flag:
            npt.assert_equal(flag, self._is_subhalo)

class TestLineOfSight(object):

    def setup_method(self):

        mass_function_model = ShethTormen
        kwargs_mass_function = {'log_mlow': 6.0, 'log_mhigh': 10.0, 'm_pivot': 1e8,
                                'LOS_normalization': 1.0, 'delta_power_law_index': 0.0}
        zlens = 0.5
        zsource = 1.5
        cosmo = Cosmology()
        geometry = Geometry(cosmo, zlens, zsource, 6.0, 'DOUBLE_CONE')
        lens_cosmo = LensCosmo(zlens, zsource, cosmo)
        lens_plane_redshifts, delta_z_list = generate_lens_plane_redshifts(zlens, zsource)
        spatial_distribution_model = LensConeUniform(geometry.cone_opening_angle, geometry)
        self.model = LineOfSight(mass_function_model, kwargs_mass_function, spatial_distribution_model,
                 geometry, lens_cosmo, lens_plane_redshifts, delta_z_list)
        self._is_subhalo = False

    def test_name(self):
        npt.assert_string_equal('LINE_OF_SIGHT', self.model.name)

    def test_render(self):

        masses, x, y, r3d, redshifts, subhalo_flag = self.model.render()
        npt.assert_equal(len(masses), len(x))
        npt.assert_equal(len(masses), len(y))
        npt.assert_equal(len(redshifts), len(masses))
        for flag in subhalo_flag:
            npt.assert_equal(flag, self._is_subhalo)

    def test_sheets(self):
        kwargs_out, profile_names_out, redshifts = self.model.convergence_sheet_correction(kappa_scale=1.0,
                                                                                           log_mlow=6.0,
                                                                                           log_mhigh=10.0)
        npt.assert_equal(len(kwargs_out), len(profile_names_out))
        npt.assert_equal(len(profile_names_out), len(redshifts))

if __name__ == '__main__':
    pytest.main()




import numpy.testing as npt
import pytest
from pyHalo.mass_function_models import preset_mass_function_models

class TestMassFunctionModels(object):

    def setup_method(self):
        self.model_list = ['SHMF_LOVELL2020', 'LOVELL2020', 'SHMF_LOVELL2014',
                           'LOVELL2014', 'SCHIVE2016', 'SHMF_SCHIVE2016', 'POWER_LAW',
                           'POWER_LAW_TURNOVER', 'POWER_LAW_TURNOVER_SHMF']
        self.model_names = ['WDM_POWER_LAW', 'SHETH_TORMEN', 'WDM_POWER_LAW',
                            'SHETH_TORMEN', 'SHETH_TORMEN', 'WDM_POWER_LAW', 'CDM_POWER_LAW',
                            'SHETH_TORMEN', 'WDM_POWER_LAW']

    def test_models(self):

        for model, model_name in zip(self.model_list, self.model_names):
            mod, kw = preset_mass_function_models(model)
            npt.assert_string_equal(model_name, mod.name)


if __name__ == '__main__':
    pytest.main()



import numpy as np
import pytest
import numpy.testing as npt
from pyHalo.Rendering.MassFunctions.stucker import stucker_suppression_params, beta_from_slope, _supp_mscale
from pyHalo.mass_function_models import preset_mass_function_models


class TestStucker(object):

    def test_load_model(self):

        kwargs_model = {'dlogT_dlogk': -2.0}
        model, kwargs_stucker = preset_mass_function_models('STUCKER', kwargs_model)
        a_wdm, b_wdm, c_wdm = kwargs_stucker['a_wdm'], kwargs_stucker['b_wdm'], kwargs_stucker['c_wdm']
        a, b, c = stucker_suppression_params(-2.0)
        npt.assert_almost_equal(a, a_wdm)
        npt.assert_almost_equal(b, b_wdm)
        npt.assert_almost_equal(c, c_wdm)

        kwargs_model = {}
        args = ('STUCKER', kwargs_model)
        npt.assert_raises(Exception, preset_mass_function_models, *args)
        kwargs_model = {'dlogT_dlogk': -2.0, 'a_wdm': 1.0}
        args = ('STUCKER', kwargs_model)
        npt.assert_raises(Exception, preset_mass_function_models, *args)
        kwargs_model = {'dlogT_dlogk': -2.0, 'b_wdm': 1.0}
        args = ('STUCKER', kwargs_model)
        npt.assert_raises(Exception, preset_mass_function_models, *args)
        kwargs_model = {'dlogT_dlogk': -2.0, 'c_wdm': 1.0}
        args = ('STUCKER', kwargs_model)
        npt.assert_raises(Exception, preset_mass_function_models, *args)

    def test_beta(self):

        beta = beta_from_slope(-1.3)
        npt.assert_almost_equal(beta, 2.0, 2)

    def test_abc_params(self):
        dlogT_dlogk = -2.0
        _a_wdm, b_wdm, c_wdm = stucker_suppression_params(dlogT_dlogk)
        a_wdm = _a_wdm ** (1/b_wdm)
        theory_mass_scale_m20 = _supp_mscale(a_wdm, b_wdm, c_wdm, frac=0.2)
        theory_mass_scale_m50 = _supp_mscale(a_wdm, b_wdm, c_wdm, frac=0.5)
        theory_mass_scale_m80 = _supp_mscale(a_wdm, b_wdm, c_wdm, frac=0.8)

        suppression_model = (1 + (a_wdm/theory_mass_scale_m20) ** b_wdm) ** c_wdm
        npt.assert_almost_equal(suppression_model, 0.2)
        suppression_model = (1 + (a_wdm/theory_mass_scale_m50) ** b_wdm) ** c_wdm
        npt.assert_almost_equal(suppression_model, 0.5)
        suppression_model = (1 + (a_wdm/theory_mass_scale_m80) ** b_wdm) ** c_wdm
        npt.assert_almost_equal(suppression_model, 0.8)

        dlogT_dlogk = -4.0
        _a_wdm, b_wdm, c_wdm = stucker_suppression_params(dlogT_dlogk)
        a_wdm = _a_wdm ** (1 / b_wdm)

        theory_mass_scale_m20 = _supp_mscale(a_wdm, b_wdm, c_wdm, frac=0.2)
        theory_mass_scale_m50 = _supp_mscale(a_wdm, b_wdm, c_wdm, frac=0.5)
        theory_mass_scale_m80 = _supp_mscale(a_wdm, b_wdm, c_wdm, frac=0.8)

        suppression_model = (1 + (a_wdm / theory_mass_scale_m20) ** b_wdm) ** c_wdm
        npt.assert_almost_equal(suppression_model, 0.2)
        suppression_model = (1 + (a_wdm / theory_mass_scale_m50) ** b_wdm) ** c_wdm
        npt.assert_almost_equal(suppression_model, 0.5)
        suppression_model = (1 + (a_wdm / theory_mass_scale_m80) ** b_wdm) ** c_wdm
        npt.assert_almost_equal(suppression_model, 0.8)

        dlogT_dlogk = -6.0
        _a_wdm, b_wdm, c_wdm = stucker_suppression_params(dlogT_dlogk)
        a_wdm = _a_wdm ** (1 / b_wdm)

        theory_mass_scale_m20 = _supp_mscale(a_wdm, b_wdm, c_wdm, frac=0.2)
        theory_mass_scale_m50 = _supp_mscale(a_wdm, b_wdm, c_wdm, frac=0.5)
        theory_mass_scale_m80 = _supp_mscale(a_wdm, b_wdm, c_wdm, frac=0.8)

        suppression_model = (1 + (a_wdm / theory_mass_scale_m20) ** b_wdm) ** c_wdm
        npt.assert_almost_equal(suppression_model, 0.2)
        suppression_model = (1 + (a_wdm / theory_mass_scale_m50) ** b_wdm) ** c_wdm
        npt.assert_almost_equal(suppression_model, 0.5)
        suppression_model = (1 + (a_wdm / theory_mass_scale_m80) ** b_wdm) ** c_wdm
        npt.assert_almost_equal(suppression_model, 0.8)


if __name__ == '__main__':
   pytest.main()



from copy import copy
from pyHalo.Cosmology.geometry import Geometry
from pyHalo.Rendering.MassFunctions.density_peaks import ShethTormen
from pyHalo.Rendering.MassFunctions.mass_function_base import CDMPowerLaw
from pyHalo.Rendering.SpatialDistributions.nfw import ProjectedNFW
from pyHalo.Rendering.SpatialDistributions.uniform import LensConeUniform
from pyHalo.concentration_models import preset_concentration_models
from pyHalo.pyhalo import pyHalo
from pyHalo.truncation_models import truncation_models


def CDM(z_lens, z_source, sigma_sub=0.025, log_mlow=6., log_mhigh=10., log10_sigma_sub=None,
        concentration_model_subhalos='DIEMERJOYCE19', kwargs_concentration_model_subhalos={},
        concentration_model_fieldhalos='DIEMERJOYCE19', kwargs_concentration_model_fieldhalos={},
        truncation_model_subhalos='TRUNCATION_GALACTICUS', kwargs_truncation_model_subhalos={},
        truncation_model_fieldhalos='TRUNCATION_RN', kwargs_truncation_model_fieldhalos={},
        shmf_log_slope=-1.9, cone_opening_angle_arcsec=6., log_m_host=13.3,  r_tidal=0.25,
        LOS_normalization=1.0, two_halo_contribution=True, delta_power_law_index=0.0,
        geometry_type='DOUBLE_CONE', kwargs_cosmo=None, host_scaling_factor=0.5,
        redshift_scaling_factor=0.3, two_halo_Lazar_correction=True, draw_poisson=True, c_host=6.0):
    """
    This class generates realizations of dark matter structure in Cold Dark Matter

    :param z_lens: main deflector redshift
    :param z_source: source redshift
    :param sigma_sub: amplitude of the subhalo mass function at 10^8 solar masses in units [# of halos / kpc^2]
    :param log10_sigma_sub: optional setting of sigma_sub in log10-scale (useful for log-uniform priors); if this is specified
    it overwrites the value of sigma_sub
    :param log_mlow: log base 10 of the minimum halo mass to render
    :param log_mhigh: log base 10 of the maximum halo mass to render
    :param concentration_model_subhalos: the concentration-mass relation applied to subhalos,
    see concentration_models.py for a complete list of available models
    :param kwargs_concentration_model_subhalos: keyword arguments for the subhalo MC relation
    NOTE: keyword args returned by the load_concentration_model override these keywords with duplicate arguments
    :param concentration_model_subhalos: the concentration-mass relation applied to field halos,
    see concentration_models.py for a complete list of available models
    :param kwargs_concentration_model_fieldhalos: keyword arguments for the field halo MC relation
    NOTE: keyword args returned by the load_concentration_model override these keywords with duplicate arguments
    :param truncation_model_subhalos: the truncation model applied to subhalos, see truncation_models for a complete list
    :param kwargs_truncation_model_subhalos: keyword arguments for the truncation model applied to subhalos
    :param truncation_model_fieldhalos: the truncation model applied to field halos, see truncation_models for a
    complete list
    :param kwargs_truncation_model_fieldhalos: keyword arguments for the truncation model applied to field halos
    :param shmf_log_slope: the logarithmic slope of the subhalo mass function pivoting around 10^8 M_sun
    :param cone_opening_angle_arcsec: the opening angle of the rendering volume in arcsec
    :param log_m_host: log base 10 of the host halo mass [M_sun]
    :param r_tidal: the core radius of the host halo in units of the host halo scale radius. Subhalos are distributed
    in 3D with a cored NFW profile with this core radius
    :param LOS_normalization: rescales the amplitude of the line-of-sight halo mass function
    :param two_halo_contribution: bool; turns on and off the two-halo term
    :param delta_power_law_index: tilts the logarithmic slope of the subhalo and field halo mass functions around pivot
    at 10^8 M_sun
    :param geometry_type: string that specifies the geometry of the rendering volume; options include
    DOUBLE_CONE, CONE, CYLINDER
    :param kwargs_cosmo: keyword arguments that specify the cosmology (see pyHalo.Cosmology.cosmology)
    :param host_scaling_factor: the scaling with host halo mass of the projected number density of subhalos
    :param redshift_scaling_factor: the scaling with (1+z) of the projected number density of subhalos
    :param two_halo_Lazar_correction: bool; if True, adds the correction to the two-halo contribution from around the
    main deflector presented by Lazar et al. (2021)
    :param c_host: manually set host halo concentration
    :return: a realization of dark matter halos
    """

    # FIRST WE CREATE AN INSTANCE OF PYHALO, WHICH SETS THE COSMOLOGY
    pyhalo = pyHalo(z_lens, z_source, kwargs_cosmo)
    # WE ALSO SPECIFY THE GEOMETRY OF THE RENDERING VOLUME
    geometry = Geometry(pyhalo.cosmology, z_lens, z_source,
                        cone_opening_angle_arcsec, geometry_type)

    # NOW WE SET THE MASS FUNCTION CLASSES FOR SUBHALOS AND FIELD HALOS
    # NOTE: MASS FUNCTION CLASSES SHOULD NOT BE INSTANTIATED HERE
    mass_function_model_subhalos = CDMPowerLaw
    mass_function_model_fieldhalos = ShethTormen

    # SET THE SPATIAL DISTRIBUTION MODELS FOR SUBHALOS AND FIELD HALOS:
    subhalo_spatial_distribution = ProjectedNFW
    fieldhalo_spatial_distribution = LensConeUniform

    # set the density profile definition
    mdef_subhalos = 'TNFW'
    mdef_field_halos = 'TNFW'

    kwargs_concentration_model_subhalos['cosmo'] = pyhalo.astropy_cosmo
    kwargs_concentration_model_fieldhalos['cosmo'] = pyhalo.astropy_cosmo

    # SET THE CONCENTRATION-MASS RELATION FOR SUBHALOS AND FIELD HALOS
    model_subhalos, kwargs_mc_subs = preset_concentration_models(concentration_model_subhalos,
                                                                 kwargs_concentration_model_subhalos)
    concentration_model_subhalos = model_subhalos(**kwargs_mc_subs)

    model_fieldhalos, kwargs_mc_field = preset_concentration_models(concentration_model_fieldhalos,
                                                                    kwargs_concentration_model_fieldhalos)
    concentration_model_fieldhalos = model_fieldhalos(**kwargs_mc_field)
    if c_host is None:
        c_host = concentration_model_fieldhalos.nfw_concentration(10 ** log_m_host, z_lens)

    # SET THE TRUNCATION RADIUS FOR SUBHALOS AND FIELD HALOS
    kwargs_truncation_model_subhalos['lens_cosmo'] = pyhalo.lens_cosmo
    kwargs_truncation_model_fieldhalos['lens_cosmo'] = pyhalo.lens_cosmo

    model_subhalos, kwargs_trunc_subs = truncation_models(truncation_model_subhalos)
    kwargs_trunc_subs.update(kwargs_truncation_model_subhalos)
    if truncation_model_subhalos in ['TRUNCATION_GALACTICUS_KEELEY24', 'TRUNCATION_GALACTICUS']:
        kwargs_trunc_subs['c_host'] = c_host
    truncation_model_subhalos = model_subhalos(**kwargs_trunc_subs)

    model_fieldhalos, kwargs_trunc_field = truncation_models(truncation_model_fieldhalos)
    kwargs_trunc_field.update(kwargs_truncation_model_fieldhalos)
    truncation_model_fieldhalos = model_fieldhalos(**kwargs_trunc_field)

    # NOW THAT THE CLASSES ARE SPECIFIED, WE SORT THE KEYWORD ARGUMENTS AND CLASSES INTO LISTS
    population_model_list = ['SUBHALOS', 'LINE_OF_SIGHT']
    mass_function_class_list = [mass_function_model_subhalos, mass_function_model_fieldhalos]
    # check for log10 value of sigma_sub
    if log10_sigma_sub is not None:
        sigma_sub = 10 ** log10_sigma_sub
    kwargs_subhalos = {'log_mlow': log_mlow,
                       'log_mhigh': log_mhigh,
                       'm_pivot': 10 ** 8,
                       'power_law_index': shmf_log_slope,
                       'delta_power_law_index': delta_power_law_index,
                       'log_m_host': log_m_host,
                       'sigma_sub': sigma_sub,
                       'host_scaling_factor': host_scaling_factor,
                       'redshift_scaling_factor': redshift_scaling_factor,
                       'draw_poisson': draw_poisson}
    kwargs_los = {'log_mlow': log_mlow,
                       'log_mhigh': log_mhigh,
                  'LOS_normalization': LOS_normalization,
                       'm_pivot': 10 ** 8,
                       'delta_power_law_index': delta_power_law_index,
                       'log_m_host': log_m_host,
                    'draw_poisson': draw_poisson}
    kwargs_mass_function_list = [kwargs_subhalos, kwargs_los]
    spatial_distribution_class_list = [subhalo_spatial_distribution, fieldhalo_spatial_distribution]
    kwargs_subhalos_spatial = {'m_host': 10 ** log_m_host, 'zlens': z_lens, 'c_host': c_host,
                               'rmax2d_arcsec': cone_opening_angle_arcsec / 2, 'r_core_units_rs': r_tidal,
                               'lens_cosmo': pyhalo.lens_cosmo}
    kwargs_los_spatial = {'cone_opening_angle': cone_opening_angle_arcsec, 'geometry': geometry}
    kwargs_spatial_distribution_list = [kwargs_subhalos_spatial, kwargs_los_spatial]

    if two_halo_contribution:
        population_model_list += ['TWO_HALO']
        kwargs_two_halo = copy(kwargs_los)
        kwargs_mass_function_list += [kwargs_two_halo]
        spatial_distribution_class_list += [LensConeUniform]
        kwargs_spatial_distribution_list += [kwargs_los_spatial]
        mass_function_class_list += [ShethTormen]

    kwargs_halo_model = {'truncation_model_subhalos': truncation_model_subhalos,
                         'concentration_model_subhalos': concentration_model_subhalos,
                         'truncation_model_field_halos': truncation_model_fieldhalos,
                         'concentration_model_field_halos': concentration_model_fieldhalos,
                         'kwargs_density_profile': {}}

    realization_list = pyhalo.render(population_model_list, mass_function_class_list, kwargs_mass_function_list,
                                     spatial_distribution_class_list, kwargs_spatial_distribution_list,
                                     geometry, mdef_subhalos, mdef_field_halos, kwargs_halo_model,
                                     two_halo_Lazar_correction,
                                     nrealizations=1)
    return realization_list[0]

def CDMCorrelatedStructure(z_lens, z_source, log_mlow=6., log_mhigh=10.,
        concentration_model='DIEMERJOYCE19', kwargs_concentration_model={},
        truncation_model='TRUNCATION_RN', kwargs_truncation_model={},
        cone_opening_angle_arcsec=6., log_m_host=13.3, LOS_normalization=1.0,
        delta_power_law_index=0.0, geometry_type='DOUBLE_CONE', kwargs_cosmo=None):
    """

    :param z_lens:
    :param z_source:
    :param log_mlow:
    :param log_mhigh:
    :param concentration_model:
    :param kwargs_concentration_model:
    :param truncation_model:
    :param kwargs_truncation_model:
    :param cone_opening_angle_arcsec:
    :param log_m_host:
    :param LOS_normalization:
    :param delta_power_law_index:
    :param geometry_type:
    :param kwargs_cosmo:
    :return:
    """
    # FIRST WE CREATE AN INSTANCE OF PYHALO, WHICH SETS THE COSMOLOGY
    pyhalo = pyHalo(z_lens, z_source, kwargs_cosmo)
    # WE ALSO SPECIFY THE GEOMETRY OF THE RENDERING VOLUME
    geometry = Geometry(pyhalo.cosmology, z_lens, z_source,
                        cone_opening_angle_arcsec, geometry_type)

    # NOW WE SET THE MASS FUNCTION CLASSES FOR SUBHALOS AND FIELD HALOS
    # NOTE: MASS FUNCTION CLASSES SHOULD NOT BE INSTANTIATED HERE
    mass_function_model = ShethTormen

    # SET THE SPATIAL DISTRIBUTION MODELS FOR SUBHALOS AND FIELD HALOS:
    fieldhalo_spatial_distribution = LensConeUniform
    kwargs_spatial_distribution = {'cone_opening_angle': cone_opening_angle_arcsec, 'geometry': geometry}

    # set the density profile definition
    mdef_subhalos = 'TNFW'
    mdef_field_halos = 'TNFW'

    kwargs_concentration_model['cosmo'] = pyhalo.astropy_cosmo

    # SET THE CONCENTRATION-MASS RELATION MODEL
    model, kwargs_mc = preset_concentration_models(concentration_model, kwargs_concentration_model)
    concentration_model = model(**kwargs_mc)

    # SET THE TRUNCATION RADIUS MODEL
    kwargs_truncation_model['lens_cosmo'] = pyhalo.lens_cosmo
    model, kwargs_trunc = truncation_models(truncation_model)
    kwargs_trunc.update(kwargs_truncation_model)
    truncation_model = model(**kwargs_trunc)

    kwargs_halo_model = {'truncation_model_subhalos': truncation_model,
                         'concentration_model_subhalos': concentration_model,
                         'truncation_model_field_halos': truncation_model,
                         'concentration_model_field_halos': concentration_model,
                         'kwargs_density_profile': {}}
    kwargs_mass_function = {'log_mlow': log_mlow,
                  'log_mhigh': log_mhigh,
                  'LOS_normalization': LOS_normalization,
                  'm_pivot': 10 ** 8,
                  'delta_power_law_index': delta_power_law_index,
                  'log_m_host': log_m_host}
    population_model_list = ['TWO_HALO']
    mass_function_class_list = [mass_function_model]
    spatial_distribution_class_list = [fieldhalo_spatial_distribution]
    kwargs_spatial_distribution_list = [kwargs_spatial_distribution]
    kwargs_mass_function_list = [kwargs_mass_function]

    realization_list = pyhalo.render(population_model_list, mass_function_class_list, kwargs_mass_function_list,
                                     spatial_distribution_class_list, kwargs_spatial_distribution_list,
                                     geometry, mdef_subhalos, mdef_field_halos, kwargs_halo_model, nrealizations=1)
    return realization_list[0]




import numpy.testing as npt
from pyHalo.Cosmology.geometry import Geometry
from pyHalo.Cosmology.cosmology import Cosmology
import pytest
import numpy as np

class TestConeGeometry(object):

    def setup_method(self):

        self.arcsec = 2 * np.pi / 360 / 3600
        self.zlens = 1
        self.zsource = 2
        self.angle_diameter = 2/self.arcsec
        self.angle_radius = 0.5*self.angle_diameter

        H0 = 70
        omega_baryon = 0.0
        omega_DM = 0.3
        sigma8 = 0.82
        curvature = 'flat'
        ns = 0.9608
        cosmo_params = {'H0': H0, 'Om0': omega_baryon + omega_DM, 'Ob0': omega_baryon,
                      'sigma8': sigma8, 'ns': ns, 'curvature': curvature}
        self.cosmo = Cosmology(cosmo_kwargs=cosmo_params)

        self.geometry_cylinder = Geometry(self.cosmo, self.zlens, self.zsource, self.angle_diameter,
                                             'CYLINDER')

    def test_angle_scale(self):

        scale = self.geometry_cylinder.rendering_scale(self.zlens)
        npt.assert_almost_equal(scale, 1.)

        scale = self.geometry_cylinder.rendering_scale(self.zlens + 0.35)
        dratio = self.geometry_cylinder.cosmo.D_C_z(self.zlens) / self.geometry_cylinder.cosmo.D_C_z(self.zlens + 0.35)
        val = 0.5 * self.geometry_cylinder.cone_opening_angle * self.arcsec * dratio
        npt.assert_almost_equal(scale, val, 3)

        dratio = self.geometry_cylinder.cosmo.D_C_z(self.zlens) / self.geometry_cylinder.cosmo.D_C_z(self.zsource)
        val = 0.5 * self.geometry_cylinder.cone_opening_angle * self.arcsec * dratio
        npt.assert_almost_equal(self.geometry_cylinder.rendering_scale(self.zsource), val)

    def test_distances_lensing(self):

        z = 0.3
        radius_physical = self.geometry_cylinder.angle_to_physicalradius(self.angle_radius, z)
        dratio = self.geometry_cylinder.cosmo.D_C_z(self.zlens) / self.geometry_cylinder.cosmo.D_C_z(z)
        radius = self.geometry_cylinder.cosmo.D_A_z(z) * self.angle_radius * self.arcsec * dratio
        npt.assert_almost_equal(radius_physical, radius, 0)

        comoving_radius = self.geometry_cylinder.angle_to_comovingradius(self.angle_radius, z)
        npt.assert_almost_equal(comoving_radius, radius_physical * (1 + z), 3)

        z = 1
        radius_physical = self.geometry_cylinder.angle_to_physicalradius(self.angle_radius, z)
        dratio = self.geometry_cylinder.cosmo.D_C_z(self.zlens) / self.geometry_cylinder.cosmo.D_C_z(z)
        radius = self.geometry_cylinder.cosmo.D_A_z(z) * self.angle_radius * self.arcsec * dratio
        npt.assert_almost_equal(radius_physical, radius, 0)

        comoving_radius = self.geometry_cylinder.angle_to_comovingradius(self.angle_radius, z)
        npt.assert_almost_equal(comoving_radius, radius_physical * (1+z), 3)

        z = 1.25
        radius_physical = self.geometry_cylinder.angle_to_physicalradius(self.angle_radius, z)
        dratio = self.geometry_cylinder.cosmo.D_C_z(self.zlens) / self.geometry_cylinder.cosmo.D_C_z(z)
        radius = self.geometry_cylinder.cosmo.D_A_z(z) * self.angle_radius * self.arcsec * dratio
        npt.assert_almost_equal(radius_physical, radius, 0)

        comoving_radius = self.geometry_cylinder.angle_to_comovingradius(self.angle_radius, z)
        npt.assert_almost_equal(comoving_radius, radius_physical * (1 + z), 3)

    def test_total_volume(self):

        delta_z = 0.001
        volume_pyhalo = 0
        for zi in np.arange(0, self.zsource+delta_z, delta_z):
            dV_pyhalo = self.geometry_cylinder.volume_element_comoving(zi, delta_z)
            volume_pyhalo += dV_pyhalo

        r = self.geometry_cylinder.cosmo.D_C_z(self.zlens) * self.geometry_cylinder.cone_opening_angle * 0.5 * self.arcsec
        d = self.geometry_cylinder.cosmo.D_C_z(self.zsource)
        volume = np.pi * r ** 2 * d
        npt.assert_almost_equal(volume/volume_pyhalo, 1, 3)

if __name__ == '__main__':
      pytest.main()



import numpy.testing as npt
import numpy as np
from pyHalo.Halos.lens_cosmo import LensCosmo
from pyHalo.Halos.HaloModels.generalized_nfw import GeneralNFWFieldHalo, GeneralNFWSubhalo
from pyHalo.Halos.HaloModels.NFW import NFWFieldHalo
from pyHalo.Halos.concentration import ConcentrationDiemerJoyce
from lenstronomy.LensModel.Profiles.pseudo_double_powerlaw import PseudoDoublePowerlaw
from lenstronomy.LensModel.Profiles.nfw import NFW
import pytest
from astropy.cosmology import FlatLambdaCDM
from pyHalo.Cosmology.cosmology import Cosmology

class TestGeneralNFW(object):

    def setup_method(self):

        astropy = FlatLambdaCDM(70.0, 0.3)
        cosmo = Cosmology(astropy_instance=astropy)
        self.zhalo = 0.5
        self.zsource = 2.0
        self.lens_cosmo = LensCosmo(self.zhalo, self.zsource, cosmo)
        self.nfw_profile_lenstronomy = NFW()
        self.gnfw_profile_lenstronomy = PseudoDoublePowerlaw()
        self.truncation_class = None
        self.concentration_class = ConcentrationDiemerJoyce(self.lens_cosmo, scatter=False)

    def test_lenstronomy_params(self):

        m = 10 ** 8
        x = 0.5
        y = 1.0
        r3d = 100
        is_subhalo = False
        gamma_inner = 1.0001
        gamma_outer = 3.0001
        x_match = 2.9
        unique_tag = 1.0
        kwargs_profile = {'gamma_inner': gamma_inner, 'gamma_outer': gamma_outer, 'x_match': x_match,
                          'evaluate_mc_at_zlens': True}
        gnfw = GeneralNFWFieldHalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, kwargs_profile,
                                   self.truncation_class, self.concentration_class, unique_tag)
        nfw = NFWFieldHalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, kwargs_profile,
                                   self.truncation_class, self.concentration_class, unique_tag)
        kwargs_nfw_profile = nfw.lenstronomy_params[0][0]
        kwargs_gnfw_profile = gnfw.lenstronomy_params[0][0]
        alpha_Rs = kwargs_gnfw_profile['alpha_Rs']
        npt.assert_almost_equal(alpha_Rs/kwargs_nfw_profile['alpha_Rs'], 1.0, 2)
        rs = kwargs_nfw_profile['Rs']
        npt.assert_almost_equal(rs/kwargs_gnfw_profile['Rs'], 1.0, 4)

        id = gnfw.lenstronomy_ID
        npt.assert_string_equal('PSEUDO_DPL', id[0])

    def test_enclosed_mass(self):

        m = 10 ** 8
        x = 0.5
        y = 1.0
        r3d = 100
        is_subhalo = False
        gamma_inner = 2.2
        gamma_outer = 3.2
        x_match = 3.5
        unique_tag = 1.0
        kwargs_profile = {'gamma_inner': gamma_inner, 'gamma_outer': gamma_outer, 'x_match': x_match,
                          'evaluate_mc_at_zlens': True}
        gnfw = GeneralNFWFieldHalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, kwargs_profile,
                                   self.truncation_class, self.concentration_class, unique_tag)
        nfw = NFWFieldHalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, kwargs_profile,
                                   self.truncation_class, self.concentration_class, unique_tag)

        kwargs_gnfw = gnfw.lenstronomy_params[0][0]
        kwargs_nfw = nfw.lenstronomy_params[0][0]
        rmatch = x_match * kwargs_nfw['Rs']
        m3d_nfw = self.nfw_profile_lenstronomy.mass_3d_lens(rmatch, kwargs_nfw['Rs'],
                                                            kwargs_nfw['alpha_Rs'])
        m3d = self.gnfw_profile_lenstronomy.mass_3d_lens(rmatch, kwargs_gnfw['Rs'],
                                                         kwargs_gnfw['alpha_Rs'],
                                                         kwargs_gnfw['gamma_inner'], kwargs_gnfw['gamma_outer'])
        npt.assert_almost_equal(m3d_nfw/m3d, 1.0, 4)
        m = 10 ** 7.3
        x = 0.5
        y = 1.0
        r3d = 100
        is_subhalo = False
        gamma_inner = 1.5
        gamma_outer = 2.89
        x_match = 4.0
        unique_tag = 1.0
        kwargs_profile = {'gamma_inner': gamma_inner, 'gamma_outer': gamma_outer, 'x_match': x_match}
        gnfw = GeneralNFWFieldHalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, kwargs_profile,
                                   self.truncation_class, self.concentration_class, unique_tag)
        nfw = NFWFieldHalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, kwargs_profile,
                           self.truncation_class, self.concentration_class, unique_tag)
        kwargs_gnfw = gnfw.lenstronomy_params[0][0]
        kwargs_nfw = nfw.lenstronomy_params[0][0]
        rmatch = x_match * kwargs_nfw['Rs']
        m3d_nfw = self.nfw_profile_lenstronomy.mass_3d_lens(rmatch, kwargs_nfw['Rs'],
                                                            kwargs_nfw['alpha_Rs'])
        m3d = self.gnfw_profile_lenstronomy.mass_3d_lens(rmatch, kwargs_gnfw['Rs'],
                                                         kwargs_gnfw['alpha_Rs'],
                                                         kwargs_gnfw['gamma_inner'], kwargs_gnfw['gamma_outer'])
        npt.assert_almost_equal(m3d_nfw / m3d, 1.0, 4)

    def test_concentration_redshift_eval(self):
        m = 10 ** 8
        x = 0.5
        y = 1.0
        r3d = 100
        gamma_inner = 1.25
        gamma_outer = 3.3
        x_match = 2.5
        unique_tag = 1.0
        kwargs_profile = {'gamma_inner': gamma_inner, 'gamma_outer': gamma_outer, 'x_match': x_match,
                          'evaluate_mc_at_zlens': False}
        is_subhalo = False
        gnfw = GeneralNFWFieldHalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, kwargs_profile,
                                   self.truncation_class, self.concentration_class, unique_tag)
        is_subhalo = True
        gnfw_subhalo = GeneralNFWSubhalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, kwargs_profile,
                                   self.truncation_class, self.concentration_class, unique_tag)
        c = gnfw.profile_args[0]
        c_sub = gnfw_subhalo.profile_args[0]
        npt.assert_equal(False, c==c_sub)

    # def test_profile(self):
    #     m = 10 ** 8
    #     x = 0.5
    #     y = 1.0
    #     r3d = 100
    #     is_subhalo = False
    #     gamma_inner = 1.0
    #     gamma_outer = 3.001
    #     x_match = 'c'
    #     unique_tag = 1.0
    #     kwargs_profile = {'gamma_inner': gamma_inner, 'gamma_outer': gamma_outer, 'x_match': x_match,
    #                       'evaluate_mc_at_zlens': True}
    #     gnfw = GeneralNFWFieldHalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, kwargs_profile,
    #                                self.truncation_class, self.concentration_class, unique_tag)
    #     nfw = NFWFieldHalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, kwargs_profile,
    #                        self.truncation_class, self.concentration_class, unique_tag)
    #     import matplotlib.pyplot as plt
    #     kwargs_gnfw = gnfw.lenstronomy_params[0][0]
    #     kwargs_nfw = nfw.lenstronomy_params[0][0]
    #
    #     rs = gnfw.nfw_params[1]
    #     x = np.logspace(-1.5, 1, 100)
    #     r = rs * x
    #     del kwargs_gnfw['center_x']
    #     del kwargs_gnfw['center_y']
    #     del kwargs_nfw['center_x']
    #     del kwargs_nfw['center_y']
    #     # plt.loglog(x, self.gnfw_profile_lenstronomy.density_lens(r, **kwargs_gnfw))
    #     # plt.loglog(x, self.nfw_profile_lenstronomy.density_lens(r, **kwargs_nfw), color='k')
    #     plt.loglog(x, gnfw.density_profile_3d(r), color='r')
    #     plt.loglog(x, nfw.density_profile_3d(r), color='k')
    #     plt.show()


if __name__ == '__main__':
    pytest.main()



import pytest
import numpy.testing as npt
from pyHalo.pyhalo import pyHalo
from pyHalo.defaults import cosmo_default
from pyHalo.Cosmology.geometry import Geometry
from pyHalo.Rendering.MassFunctions.mass_function_base import CDMPowerLaw
from pyHalo.Rendering.MassFunctions.density_peaks import ShethTormen
from pyHalo.Rendering.SpatialDistributions.uniform import LensConeUniform
from pyHalo.Rendering.SpatialDistributions.nfw import ProjectedNFW
from pyHalo.Halos.tidal_truncation import TruncationRN, TruncationRoche
from pyHalo.Halos.concentration import ConcentrationDiemerJoyce

class TestpyHalo(object):

    def setup_method(self):
        pass
        #self.pyhalo = pyHalo(0.5, 2.)

    def test_create_realizations(self):
        pyhalo = pyHalo(0.5, 2.)
        cone_opening_angle = 6.0
        geometry = Geometry(pyhalo.cosmology, pyhalo.zlens, pyhalo.zsource, cone_opening_angle, 'DOUBLE_CONE')
        population_model_list = ['SUBHALOS', 'LINE_OF_SIGHT', 'LINE_OF_SIGHT_NOSHEET', 'TWO_HALO']
        mass_function_class_list = [CDMPowerLaw, ShethTormen, ShethTormen, ShethTormen]
        kwargs_subhalos = {'log_mlow': 6.0, 'log_mhigh': 10.0, 'm_pivot': 10**8, 'power_law_index': -1.9,
                           'delta_power_law_index': 0.0, 'draw_poisson': False, 'log_m_host': 13.3,
                           'sigma_sub': 0.05, 'host_scaling_factor': 0.88, 'redshift_scaling_factor': 1.7}
        kwargs_los = {'log_mlow': 6.0, 'log_mhigh': 10.0, 'm_pivot': 1e8,
                      'LOS_normalization': 1.0, 'delta_power_law_index': 0.0}
        kwargs_two_halo = {'log_mlow': 6.0, 'log_mhigh': 10.0, 'm_pivot': 1e8,
                           'LOS_normalization': 1.0, 'delta_power_law_index': 0.0, 'log_m_host': 13.0}
        kwargs_mass_function_list = [kwargs_subhalos, kwargs_los, kwargs_los, kwargs_two_halo]
        spatial_distribution_class_list = [ProjectedNFW, LensConeUniform, LensConeUniform, LensConeUniform]
        kwargs_subhalos_spatial = {'m_host': 10 ** 13.0, 'zlens': 0.5, 'rmax2d_arcsec': 3.0, 'r_core_units_rs': 0.2,
                                   'lens_cosmo': pyhalo.lens_cosmo}
        kwargs_los_spatial = {'cone_opening_angle': 6.0, 'geometry': geometry}
        kwargs_spatial_distribution_list = [kwargs_subhalos_spatial, kwargs_los_spatial,
                                       kwargs_los_spatial, kwargs_los_spatial]
        mdef_subhalos = 'TNFW'
        mdef_field_halos = 'TNFW'
        kwargs_halo_model = {'truncation_model_subhalos': TruncationRoche(pyhalo.lens_cosmo),
                             'concentration_model_subhalos': ConcentrationDiemerJoyce(pyhalo.lens_cosmo),
                             'truncation_model_field_halos': TruncationRN(pyhalo.lens_cosmo),
                             'concentration_model_field_halos': ConcentrationDiemerJoyce(pyhalo.lens_cosmo),
                             'kwargs_density_profile': {}}
        realization_list = pyhalo.render(population_model_list,
                                         mass_function_class_list,
                                         kwargs_mass_function_list,
                                         spatial_distribution_class_list,
                                         kwargs_spatial_distribution_list,
               geometry, mdef_subhalos, mdef_field_halos, kwargs_halo_model, nrealizations=2)
        npt.assert_equal(len(realization_list), 2)

    def test_cosmology_setup_from_default(self):
        pyhalo = pyHalo(0.5, 2.)
        astropy = pyhalo.astropy_cosmo
        npt.assert_equal(astropy.h * 100, cosmo_default('H0'))
        npt.assert_equal(astropy.Om0, cosmo_default('Om0'))
        npt.assert_equal(astropy.Ob0, cosmo_default('Ob0'))
        if cosmo_default.cosmo_param_dictionary['flat']:
            npt.assert_equal(astropy.Om0 + astropy.Ode0, 1.0)
        else:
            npt.assert_equal(False, astropy.Om0 + astropy.Ode0!=1.0)

    def test_cosmology_setup_from_kwargs(self):

        kwargs_cosmo = {'H0': 70.0,
                        'Ob0': 0.05,
                        'Om0': 0.25,
                        'sigma8': 0.81,
                        'flat': True,
                        'ns': 0.965,
                        'power_law': False}

        pyhalo = pyHalo(0.5, 2.0, kwargs_cosmo)
        astropy = pyhalo.astropy_cosmo
        npt.assert_equal(astropy.h * 100, kwargs_cosmo['H0'])
        npt.assert_equal(astropy.Om0, kwargs_cosmo['Om0'])
        npt.assert_equal(astropy.Ob0, kwargs_cosmo['Ob0'])
        npt.assert_equal(astropy.Om0 + astropy.Ode0, 1.0)
        colosus_cosmo = pyhalo.cosmology.colossus
        npt.assert_equal(colosus_cosmo.flat, kwargs_cosmo['flat'])
        npt.assert_equal(colosus_cosmo.Om0, kwargs_cosmo['Om0'])
        npt.assert_equal(colosus_cosmo.Ob0, kwargs_cosmo['Ob0'])
        npt.assert_equal(colosus_cosmo.sigma8, kwargs_cosmo['sigma8'])
        npt.assert_equal(colosus_cosmo.ns, kwargs_cosmo['ns'])

    def test_cosmology_dependence(self):

        # make sure that boosting stuff like total amount of dark matter actually increases the number of halos
        mdef_subhalos = 'TNFW'
        mdef_field_halos = 'TNFW'
        population_model_list = ['LINE_OF_SIGHT']
        kwargs_los = {'log_mlow': 6.0, 'log_mhigh': 10.0, 'm_pivot': 1e8,
                      'LOS_normalization': 1.0, 'delta_power_law_index': 0.0}
        kwargs_mass_function_list = [kwargs_los]
        spatial_distribution_class_list = [LensConeUniform]
        mass_function_class_list = [ShethTormen]
        kwargs_cosmo_1 = {'H0': 70.0,
                        'Ob0': 0.05,
                        'Om0': 0.2,
                        'sigma8': 1.0,
                        'flat': True,
                        'ns': 0.965,
                        'power_law': False}
        pyhalo_1 = pyHalo(0.5, 2.0, kwargs_cosmo_1)
        cone_opening_angle = 6.0
        geometry_1 = Geometry(pyhalo_1.cosmology, pyhalo_1.zlens, pyhalo_1.zsource, cone_opening_angle,
                              'DOUBLE_CONE')

        kwargs_spatial_distribution_list_1 = [{'cone_opening_angle': 6.0, 'geometry': geometry_1}]
        kwargs_halo_model_1 = {'truncation_model_subhalos': TruncationRoche(pyhalo_1.lens_cosmo),
                               'concentration_model_subhalos': ConcentrationDiemerJoyce(pyhalo_1.lens_cosmo),
                               'truncation_model_field_halos': TruncationRN(pyhalo_1.lens_cosmo),
                               'concentration_model_field_halos': ConcentrationDiemerJoyce(pyhalo_1.lens_cosmo),
                               'kwargs_density_profile': {}}
        realization_list_1 = pyhalo_1.render(population_model_list, mass_function_class_list,
                                             kwargs_mass_function_list,
                                             spatial_distribution_class_list,
                                             kwargs_spatial_distribution_list_1,
                                             geometry_1, mdef_subhalos, mdef_field_halos, kwargs_halo_model_1,
                                             nrealizations=1)
        nhalos_1 = len(realization_list_1[0].halos)


        kwargs_cosmo_2 = {'H0': 70.0,
                          'Ob0': 0.05,
                          'Om0': 0.8,
                          'sigma8': 1.0,
                          'flat': True,
                          'ns': 0.965,
                          'power_law': False}
        pyhalo_2 = pyHalo(0.5, 2.0, kwargs_cosmo_2)
        geometry_2 = Geometry(pyhalo_2.cosmology, pyhalo_2.zlens, pyhalo_2.zsource, cone_opening_angle,
                              'DOUBLE_CONE')
        kwargs_spatial_distribution_list_2 = [{'cone_opening_angle': 6.0, 'geometry': geometry_2}]
        kwargs_halo_model_2 = {'truncation_model_subhalos': TruncationRoche(pyhalo_2.lens_cosmo),
                             'concentration_model_subhalos': ConcentrationDiemerJoyce(pyhalo_2.astropy_cosmo),
                             'truncation_model_field_halos': TruncationRN(pyhalo_2.lens_cosmo),
                             'concentration_model_field_halos': ConcentrationDiemerJoyce(pyhalo_2.astropy_cosmo),
                             'kwargs_density_profile': {}}
        realization_list_2 = pyhalo_2.render(population_model_list, mass_function_class_list,
                                             kwargs_mass_function_list,
                                             spatial_distribution_class_list,
                                             kwargs_spatial_distribution_list_2,
                                             geometry_2, mdef_subhalos, mdef_field_halos, kwargs_halo_model_2,
                                             nrealizations=1)
        nhalos_2 = len(realization_list_2[0].halos)
        npt.assert_array_less(nhalos_1, nhalos_2)


if __name__ == '__main__':

    pytest.main()



import numpy
from scipy.interpolate import interp1d
from scipy.optimize import minimize
from lenstronomy.Cosmo.nfw_param import NFWParam
import astropy.units as un
from colossus.lss.bias import twoHaloTerm
from scipy.integrate import quad

class NFWParampyHalo(NFWParam):

    """Adds methods for psuedo-NFW profiles to those implemented in lenstronomy"""

    def rho0_c_pseudoNFW(self, c, z):
        """
        computes density normalization as a function of concentration parameter for an NFW-like profile
        with (1+x^2) in the denominator instead of (1+x)^2

        :param c: concentration
        :param z: redshift
        :return: density normalization in h^2/Mpc^3 (physical)
        """
        return 2 * c ** 3 * self.rhoc_z(z) * 200 / (3 * numpy.log(1+c**2))


class LensCosmo(object):

    def __init__(self, z_lens=None, z_source=None, cosmology=None):
        """

        This class performs calcuations relevant for certain halo mass profiles and lensing-related quantities for a
        given lens/source redshift and cosmology
        :param z_lens: deflector redshift
        :param z_source: source redshift
        :param cosmology: and instance of the Cosmology class (see pyhalo.Cosmology.cosmology.py)
        """
        if cosmology is None:
            from pyHalo.Cosmology.cosmology import Cosmology
            cosmology = Cosmology()

        self.cosmo = cosmology
        self._arcsec = 2 * numpy.pi / 360 / 3600
        self.h = self.cosmo.h
        # critical density of the universe in M_sun h^2 Mpc^-3
        rhoc = un.Quantity(self.cosmo.astropy.critical_density(0), unit=un.Msun / un.Mpc ** 3).value
        self.rhoc = rhoc / self.cosmo.h ** 2

        if z_lens is not None and z_source is not None:
            # critical density for lensing in units M_sun * Mpc ^ -2
            self.sigma_crit_lensing = self.get_sigma_crit_lensing(z_lens, z_source)
            # critical density for lensing in units M_sun * kpc ^ -2
            self.sigma_crit_lens_kpc = self.sigma_crit_lensing * (0.001) ** 2
            # critical density for lensing in units M_sun * arcsec ^ -2 at lens redshift
            self.sigmacrit = self.sigma_crit_lensing * (0.001) ** 2 * self.cosmo.kpc_proper_per_asec(z_lens) ** 2
            # lensing distances
            self.D_d, self.D_s, self.D_ds = self.cosmo.D_A_z(z_lens), self.cosmo.D_A_z(z_source), self.cosmo.D_A(
                z_lens, z_source)
        self._computed_zacc_pdf = False
        self._nfw_param = NFWParampyHalo(self.cosmo.astropy)
        self.z_lens = z_lens
        self.z_source = z_source
        self._z_infall_pdf = InfallDistributionGalacticus2024(z_lens)

    def z_accreted_from_zlens(self, mass, z_lens):
        """
        Returns the redshift a subhalo was accreted. Note that in the current implementation this is
        independent of infall mass
        :param mass: subhalo mass at infall
        :param z_lens: main deflector redshift
        :return: accretion redshift
        """
        return self._z_infall_pdf.z_accreted_from_zlens(z_lens)

    def two_halo_boost(self, m200, z, rmin=0.5, rmax=10):

        """
        Computes the average contribution of the two halo term in a redshift slice adjacent
        the main deflector. Returns a rescaling factor applied to the mass function normalization

        :param m200: host halo mass
        :param z: redshift
        :param rmin: lower limit of the integral, something like the virial radius ~500 kpc
        :param rmax: Upper limit of the integral, this is computed based on redshift spacing during
        the rendering of halos
        :return: scaling factor applied to the normalization of the LOS mass function
        """

        mean_boost = 2 * quad(self.twohaloterm, rmin, rmax, args=(m200, z))[0] / (rmax - rmin)
        # factor of two for symmetry in front/behind host halo

        return 1. + mean_boost

    def twohaloterm(self, r, M, z, mdef='200c'):

        """
        Computes the boost to the background density of the Universe
        from correlated structure around a host of mass M
        :param r:
        :param M:
        :param z:
        :param mdef:
        :return:
        """

        h = self.cosmo.h
        M_h = M * h
        r_h = r * h
        rho_2h = twoHaloTerm(r_h, M_h, z, mdef=mdef) / self.cosmo._colossus_cosmo.rho_m(z)
        return rho_2h

    def nfw_physical2angle(self, m, c, z, pseudo_nfw=False):
        """
        converts the physical mass and concentration parameter of an NFW profile into the lensing quantities
        updates lenstronomy implementation with arbitrary redshift

        :param m: mass enclosed 200 rho_crit in units of M_sun (physical units, meaning no little h)
        :param c: NFW concentration parameter (r200/r_s)
        :param z: redshift at which to evaluate angles from distances
        :param pseudo_nfw: specifies whether one deals with a regualr NFW profile (False) or a psuedo-NFW profile
        with (1+x^2) in the denominator rather than (1+x)^2
        :return: Rs_angle (angle at scale radius) (in units of arcsec), alpha_Rs (observed bending angle at the scale radius
        """
        rho0, Rs, _ = self.nfwParam_physical(m, c, z, pseudo_nfw)
        return self.nfw_physical2angle_fromNFWparams(rho0, Rs, z, pseudo_nfw)

    def nfw_physical2angle_fromNFWparams(self, rho0, Rs, z, pseudo_nfw=False):
        """
        computes the angular lensing properties of an NFW profile from its physical parameters
        :param rho0: central density normalization [M_sun / Mpc^3]
        :param Rs: scale radius [Mpc]
        :param z: redshift at which to evaluate angles from distances
        :param pseudo_nfw: specifies whether one deals with a regualr NFW profile (False) or a psuedo-NFW profile
        with (1+x^2) in the denominator rather than (1+x)^2
        :return: scale radius and deflection angle at the scale radius in arcsec
        """

        dd = self.cosmo.D_A_z(z)
        Rs_angle = Rs / dd / self._arcsec  # Rs in arcsec
        if pseudo_nfw:
            r2 = numpy.sqrt(2)
            alpha_Rs = rho0 * (4 * Rs ** 2 * (r2*numpy.log(1+r2) + numpy.log(1. / 2.)))
        else:
            alpha_Rs = rho0 * (4 * Rs ** 2 * (1 + numpy.log(1. / 2.)))
        sigma_crit = self.get_sigma_crit_lensing(z, self.z_source)
        return Rs_angle, alpha_Rs / sigma_crit / dd / self._arcsec

    def rN_M(self, M, z, N):
        """
        computes the radius R_N of a halo of mass M in physical mass M/h, where N is a number multiplying the critical
        density of the Universe at z

        :param M: halo mass in M_sun/h
        :param z: redshift
        :param N: number, e.g. N=200 computes r200
        :return: radius R_N in physical Mpc/h
        """
        rn_mpc_over_h = (3 * M / (4 * numpy.pi * self._nfw_param.rhoc_z(z) * N)) ** (1. / 3.)
        return rn_mpc_over_h / self.cosmo.h

    def nfwParam_physical(self, m, c, z, pseudo_nfw=False):
        """
        returns the NFW parameters in physical units
        updates lenstronomy implementation with arbitrary redshift

        :param m: physical mass in M_sun
        :param c: concentration
        :return: rho0 [Msun/Mpc^3], Rs [Mpc], r200 [Mpc]
        """
        r200 = self._nfw_param.r200_M(m * self.h, z) / self.h  # physical radius r200
        if pseudo_nfw:
            rho0 = self._nfw_param.rho0_c_pseudoNFW(c, z) * self.h ** 2 # physical density in M_sun/Mpc**3
        else:
            rho0 = self._nfw_param.rho0_c(c, z) * self.h**2  # physical density in M_sun/Mpc**3
        Rs = r200/c
        return rho0, Rs, r200

    def NFW_params_physical(self, m, c, z, pseudo_nfw=False):
        """
        returns the NFW parameters in physical units

        :param M: physical mass in M_sun
        :param c: concentration
        :param z: redshift
        :param pseudo_nfw: bool; if False, uses a regular NFW profile, if True, uses an NFW profile
        with (1+x^2) in the denominator
        :return: rho0 [Msun/kpc^3], Rs [kpc], r200 [kpc]
        """
        rho0, Rs, r200 = self.nfwParam_physical(m, c, z, pseudo_nfw)
        return rho0 * 1000 ** -3, Rs * 1000, r200 * 1000

    def sigma_crit_mass(self, z, area):

        """
        :param z: redshift
        :param area: physical area in Mpc^2
        :return: the 'critical density mass' sigma_crit * A in units M_sun
        """

        sigma_crit_mpc = self.get_sigma_crit_lensing(z, self.z_source)
        return area * sigma_crit_mpc

    @property
    def colossus(self):
        return self.cosmo.colossus

    ######################################################
    """ACCESS ROUTINES IN STRUCTURAL PARAMETERS CLASS"""
    ######################################################

    def mthermal_to_halfmode(self, m_thermal):
        """
        Convert thermal relic particle mass to half-mode mass
        :param m_thermaal:
        :return:
        """
        # too lazy for algebra
        def _func(m):
            return abs(self.halfmode_to_thermal(m)-m_thermal)/0.01
        return minimize(_func, x0=10**8, method='Nelder-Mead')['x']

    def halfmode_to_thermal(self, m_half_mode):

        """
        Converts a half mode mass in units of solar masses (no little h) to the mass of
        the corresponding thermal relic particle in keV
        :param m: half mode mass in solar masses
        :return: thermal relic particle mass in keV
        """

        omega_matter = self.cosmo.astropy.Om0
        return 2.32 * (omega_matter / 0.25)**0.4 * (self.cosmo.h/0.7)**0.8 * \
               (m_half_mode / 10 ** 9) ** (-0.3)

    def mhm_to_fsl(self, m_hm):
        """
        Converts half mode mass to free streaming length in Mpc
        See Equations 5-8 in https://arxiv.org/pdf/1112.0330.pdf
        :param m_hm: half-mode mass in units M_sun (no little h)
        :return: free streaming length in units Mpc
        """

        rhoc = self.rhoc * self.cosmo.h ** 2
        l_hm = 2 * (3 * m_hm / (4 * numpy.pi * rhoc)) ** (1. / 3)
        l_fs = l_hm / 13.93
        return l_fs

    ##################################################################################
    """ROUTINES RELATED TO LENSING STUFF"""
    ##################################################################################

    def get_sigma_crit_lensing(self, z1, z2):
        """
        Computes thee critial density for lensing in units of M_sun / Mpc^2
        :param z1: the lens redshit
        :param z2: the source redshift
        :return: the critial density for lensing
        """
        D_ds = self.cosmo.D_A(z1, z2)
        D_d = self.cosmo.D_A_z(z1)
        D_s = self.cosmo.D_A_z(z2)
        d_inv = D_s*D_ds**-1*D_d**-1
        # (Mpc ^2 / sec^2) * (Mpc^-3 M_sun^1 sec ^ 2) * Mpc ^-1 = M_sun / Mpc ^2
        epsilon_crit = (self.cosmo.c**2*(4*numpy.pi*self.cosmo.G)**-1)*d_inv
        return epsilon_crit

    def point_mass_factor_z(self, z):

        """
        Returns the cosmology-dependent factor to evaluate the Einstein radius of a point mass of mass M:

        :param z: redshift
        :return: The factor that when multiplied by sqrt(mass) gives the Einstein radius of a point mass

        R_ein = sqrt(M) * point_mass_factor_z(z)

        """
        factor = 4 * self.cosmo.G * self.cosmo.c ** -2
        dds = self.cosmo.D_A(z, self.z_source)
        dd = self.cosmo.D_A_z(z)
        ds = self.D_s
        factor *= dds / dd / ds
        return factor ** 0.5 / self.cosmo.arcsec

    def halo_dynamical_time(self, m_host, z, c_host):
        """
        This routine computes the dynamical timescale for a halo of mass M defined as
        t = 0.5427 / sqrt(G*rho)
        where G is the gravitational constant and rho is the average density
        :param m_host: host mass in M_sun
        :param z: host redshift
        :param c_host: host halo concentration
        :return: the dynamical timescale in Gyr
        """

        _, _, rvir = self.NFW_params_physical(m_host, c_host, z)
        volume = (4/3)*numpy.pi*rvir**3
        rho_average = m_host / volume
        g = 4.3e-6
        return 0.5427 / numpy.sqrt(g*rho_average)

class InfallDistributionGalacticus2024(object):
    """ACCRETION REDSHIFT PDF FROM GALACTICUS USING THE VERSION OF GALACTICUS AS OF FEB 2024 AND SELECTING ON
    SUBHALOS WITH A BOUND MASS > 10^6"""

    def __init__(self, z_lens):
        self.z_lens = z_lens
        self._counts = numpy.array([ 50,  93, 125, 180, 175, 144, 120, 117,  97,  82,  52,  51,  35,
        20,   9,   4,   4,   0,   1,   1])
        self._z_infall = numpy.array([ 0.53836189,  1.37376234,  2.2091628 ,  3.04456325,  3.87996371,
        4.71536416,  5.55076461,  6.38616507,  7.22156552,  8.05696598,
        8.89236643,  9.72776689, 10.56316734, 11.39856779, 12.23396825,
       13.0693687 , 13.90476916, 14.74016961, 15.57557007, 16.41097052]) - 0.5
        cdf = numpy.cumsum(self._counts)
        self._cdf = cdf / numpy.max(cdf)
        self._cdf_min = numpy.min(self._cdf)
        self._cdf_max = numpy.max(self._cdf)
        self._interp = interp1d(self._cdf, self._z_infall)

    def z_accreted_from_zlens(self, z_lens):
        u = numpy.random.uniform(self._cdf_min, self._cdf_max)
        z_infall = z_lens + self._interp(u)
        return z_infall


# class InfallDistributionGalacticus2024(object):
#     """ACCRETION REDSHIFT PDF FROM GALACTICUS USING THE VERSION OF GALACTICUS AS OF FEB 2024"""
#
#     def __init__(self, z_lens):
#         self.z_lens = z_lens
#         self._counts = numpy.array([ 74, 111, 138, 225, 281, 394, 396, 492, 603, 665, 626, 738, 714,
#         725, 744, 712, 679, 600, 556, 524, 478, 442, 347, 322, 283, 198,
#         189, 148, 137,  98,  77,  44,  32,  32,  26,  18,  15,   6,   5,
#         4,   0,   2,   0,   0])
#         self._z_infall = numpy.array([ 0.25,  0.75,  1.25,  1.75,  2.25,  2.75,  3.25,  3.75,  4.25,
#         4.75,  5.25,  5.75,  6.25,  6.75,  7.25,  7.75,  8.25,  8.75,
#         9.25,  9.75, 10.25, 10.75, 11.25, 11.75, 12.25, 12.75, 13.25,
#         13.75, 14.25, 14.75, 15.25, 15.75, 16.25, 16.75, 17.25, 17.75,
#         18.25, 18.75, 19.25, 19.75, 20.25, 20.75, 21.25, 21.75])
#         cdf = numpy.cumsum(self._counts)
#         self._cdf = cdf / numpy.max(cdf)
#         self._cdf_min = numpy.min(self._cdf)
#         self._cdf_max = numpy.max(self._cdf)
#         self._interp = interp1d(self._cdf, self._z_infall)
#
#     def z_accreted_from_zlens(self, z_lens):
#         u = numpy.random.uniform(self._cdf_min, self._cdf_max)
#         z_infall = z_lens + self._interp(u)
#         return z_infall
#
# class InfallDistributionGalacticus2020(object):
#     """ACCRETION REDSHIFT PDF FROM GALACTICUS USING THE VERSION OF GALACTICUS PUBLISHED IN 2020 WITH
#     WARM DARK MATTER CHILLS OUT"""
#     def __init__(self, z_lens):
#         self.z_lens = z_lens
#
#     @property
#     def _subhalo_accretion_pdfs(self):
#
#         if self._computed_zacc_pdf is False:
#             self._computed_zacc_pdf = True
#             self._mlist, self._dzvals, self._cdfs = self._Msub_cdfs(self.z_lens)
#
#         return self._mlist, self._dzvals, self._cdfs
#
#     def z_accreted_from_zlens(self, msub, zlens):
#
#         mlist, dzvals, cdfs = self._subhalo_accretion_pdfs
#
#         idx = self._mass_index(msub, mlist)
#
#         z_accreted = zlens + self._sample_cdf_single(cdfs[idx])
#
#         return z_accreted
#
#     def _cdf_numerical(self, m, z_lens, delta_z_values):
#
#         c_d_f = []
#
#         prob = 0
#         for zi in delta_z_values:
#             prob += self._P_fit_diff_M_sub(z_lens + zi, z_lens, m)
#             c_d_f.append(prob)
#         return numpy.array(c_d_f) / c_d_f[-1]
#
#     def _Msub_cdfs(self, z_lens):
#
#         M_sub_exp = numpy.arange(6.0, 10.2, 0.2)
#         M_sub_list = 10 ** M_sub_exp
#         delta_z = numpy.linspace(0., 6, 8000)
#         funcs = []
#
#         for mi in M_sub_list:
#             # cdfi = P_fit_diff_M_sub_cumulative(z_lens+delta_z, z_lens, mi)
#             cdfi = self._cdf_numerical(mi, z_lens, delta_z)
#
#             funcs.append(interp1d(cdfi, delta_z))
#
#         return M_sub_list, delta_z, funcs
#
#     def z_decay_mass_dependence(self, M_sub):
#         # Mass dependence of z_decay.
#         a = 3.21509397
#         b = 1.04659814e-03
#
#         return a - b * numpy.log(M_sub / 1.0e6) ** 3
#
#     def z_decay_exp_mass_dependence(self, M_sub):
#         # Mass dependence of z_decay_exp.
#
#         a = 0.30335749
#         b = 3.2777e-4
#
#         return a - b * numpy.log(M_sub / 1.0e6) ** 3
#
#     def _P_fit_diff_M_sub(self, z, z_lens, M_sub):
#         # Given the redhsift of the lens, z_lens, and the subhalo mass, M_sub, return the
#         # posibility that the subhlao has an accretion redhisft of z.
#
#         z_decay = self.z_decay_mass_dependence(M_sub)
#         z_decay_exp = self.z_decay_exp_mass_dependence(M_sub)
#
#         normalization = 2.0 / numpy.sqrt(2.0 * numpy.pi) / z_decay \
#                         / numpy.exp(0.5 * z_decay ** 2 * z_decay_exp ** 2) \
#                         / erfc(z_decay * z_decay_exp / numpy.sqrt(2.0))
#         return normalization * numpy.exp(-0.5 * ((z - z_lens) / z_decay) ** 2) \
#                * numpy.exp(-z_decay_exp * (z - z_lens))
#
#     def _sample_cdf_single(self, cdf_interp):
#
#         u = numpy.random.uniform(0, 1)
#
#         try:
#             output = float(cdf_interp(u))
#             if numpy.isnan(output):
#                 output = 0
#
#         except:
#             output = 0
#
#         return output
#
#     def _mass_index(self, subhalo_mass, mass_array):
#
#         idx = numpy.argmin(numpy.absolute(subhalo_mass - mass_array))
#         return idx

